#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Farid Rashidi Mehrabadi All rights reserved.

# =========================================================================================
# Author     : Farid Rashidi Mehrabadi (farid.rashidimehrabadi@nih.gov)
# Last Update: Aug 15, 2020
# Description: extracting gene expression matrix and genotype matrix
# =========================================================================================

import os
import shutil

import pandas as pd

import trisicell as tsc
from trisicell.ul._servers import *


def after02(config):
    def expression():
        files = [
            f"{config['outdir']}/{s}/expr.genes.results" for s in config["samples"]
        ]
        kinds = {"expected_count": "count", "TPM": "tpm", "FPKM": "fpkm"}
        for kind in kinds.keys():
            i = 0
            expr = pd.DataFrame()
            for file in files:
                df = pd.read_csv(file, sep="\t")
                name = os.path.split(os.path.dirname(file))[-1]
                df = df[["gene_id", "expected_count", "TPM", "FPKM"]]
                df = df.set_index(["gene_id"])
                if i == 0:
                    expr = pd.DataFrame(index=df.index)
                expr[name] = df[kind]
                i += 1
            expr = expr.transpose()
            expr = expr.sort_index(ascending=True)
            expr.to_csv(f"{config['outdir']}/_output/expr.{kinds[kind]}.tsv", sep="\t")

    def genotype():
        files = [
            f"{config['outdir']}/_annotating/annovar_final/{s}.txt"
            for s in config["samples"]
        ]
        data = []
        for file in files:
            try:
                df = pd.read_csv(file, sep="\t", header=None)
                data.append(df)
            except:
                print(file)
        df = data[0]
        for df2 in data[1:]:
            df = df.append(df2, ignore_index=True)
        df = df.drop_duplicates()
        df.to_csv(
            f"{config['outdir']}/_output/exonic.tsv", index=None, sep="\t", header=None
        )
        shutil.copyfile(
            f"{config['outdir']}/_calling/jointcalls.g.vcf",
            f"{config['outdir']}/_output/jointcalls.g.vcf",
        )

    def readstats():
        def get_numreads_percmapped(file):
            data = {
                "uniquely_mapped_percent": None,
                "num_splices": None,
                "num_GCAG_splices": None,
                "insertion_length": None,
                "deletion_length": None,
                "unmapped_tooshort_percent": None,
                "avg_mapped_read_length": None,
                "deletion_rate": None,
                "mismatch_rate": None,
                "avg_input_read_length": None,
                "num_ATAC_splices": None,
                "num_annotated_splices": None,
                "num_GTAG_splices": None,
                "uniquely_mapped": None,
                "multimapped_toomany": None,
                "unmapped_mismatches": None,
                "unmapped_mismatches_percent": None,
                "total_reads": None,
                "unmapped_other": None,
                "insertion_rate": None,
                "unmapped_other_percent": None,
                "multimapped_percent": None,
                "multimapped": None,
                "num_noncanonical_splices": None,
                "unmapped_tooshort": None,
                "multimapped_toomany_percent": None,
            }
            with open(file) as fin:
                for line in fin:
                    if "Uniquely mapped reads % |" in line:
                        data["uniquely_mapped_percent"] = float(
                            line.replace("Uniquely mapped reads % |", "")
                            .strip()
                            .replace("%", "")
                        )
                    if "Number of splices: Total |" in line:
                        data["num_splices"] = int(
                            line.replace("Number of splices: Total |", "").strip()
                        )
                    if "Number of splices: GC/AG |" in line:
                        data["num_GCAG_splices"] = int(
                            line.replace("Number of splices: GC/AG |", "").strip()
                        )
                    if "Insertion average length |" in line:
                        data["insertion_length"] = float(
                            line.replace("Insertion average length |", "")
                            .strip()
                            .replace("%", "")
                        )
                    if "Deletion average length |" in line:
                        data["deletion_length"] = float(
                            line.replace("Deletion average length |", "")
                            .strip()
                            .replace("%", "")
                        )
                    if "% of reads unmapped: too short |" in line:
                        data["unmapped_tooshort_percent"] = float(
                            line.replace("% of reads unmapped: too short |", "")
                            .strip()
                            .replace("%", "")
                        )
                    if "Average mapped length |" in line:
                        data["avg_mapped_read_length"] = float(
                            line.replace("Average mapped length |", "")
                            .strip()
                            .replace("%", "")
                        )
                    if "Deletion rate per base |" in line:
                        data["deletion_rate"] = float(
                            line.replace("Deletion rate per base |", "")
                            .strip()
                            .replace("%", "")
                        )
                    if "Mismatch rate per base, % |" in line:
                        data["mismatch_rate"] = float(
                            line.replace("Mismatch rate per base, % |", "")
                            .strip()
                            .replace("%", "")
                        )
                    if "Average input read length |" in line:
                        data["avg_input_read_length"] = float(
                            line.replace("Average input read length |", "")
                            .strip()
                            .replace("%", "")
                        )
                    if "Number of splices: AT/AC |" in line:
                        data["num_ATAC_splices"] = int(
                            line.replace("Number of splices: AT/AC |", "").strip()
                        )
                    if "Number of splices: Annotated (sjdb) |" in line:
                        data["num_annotated_splices"] = int(
                            line.replace(
                                "Number of splices: Annotated (sjdb) |", ""
                            ).strip()
                        )
                    if "Number of splices: GT/AG |" in line:
                        data["num_GTAG_splices"] = int(
                            line.replace("Number of splices: GT/AG |", "").strip()
                        )
                    if "Uniquely mapped reads number |" in line:
                        data["uniquely_mapped"] = int(
                            line.replace("Uniquely mapped reads number |", "").strip()
                        )
                    if "Number of reads mapped to too many loci |" in line:
                        data["multimapped_toomany"] = int(
                            line.replace(
                                "Number of reads mapped to too many loci |", ""
                            ).strip()
                        )
                    if "Number of reads unmapped: too many mismatches |" in line:
                        data["unmapped_mismatches"] = int(
                            line.replace(
                                "Number of reads unmapped: too many mismatches |", ""
                            ).strip()
                        )
                    if "% of reads unmapped: too many mismatches |" in line:
                        data["unmapped_mismatches_percent"] = float(
                            line.replace(
                                "% of reads unmapped: too many mismatches |", ""
                            )
                            .strip()
                            .replace("%", "")
                        )
                    if "Number of input reads |" in line:
                        data["total_reads"] = int(
                            line.replace("Number of input reads |", "").strip()
                        )
                    if "Number of reads unmapped: other |" in line:
                        data["unmapped_other"] = int(
                            line.replace(
                                "Number of reads unmapped: other |", ""
                            ).strip()
                        )
                    if "Insertion rate per base |" in line:
                        data["insertion_rate"] = float(
                            line.replace("Insertion rate per base |", "")
                            .strip()
                            .replace("%", "")
                        )
                    if "% of reads unmapped: other |" in line:
                        data["unmapped_other_percent"] = float(
                            line.replace("% of reads unmapped: other |", "")
                            .strip()
                            .replace("%", "")
                        )
                    if "% of reads mapped to multiple loci |" in line:
                        data["multimapped_percent"] = float(
                            line.replace("% of reads mapped to multiple loci |", "")
                            .strip()
                            .replace("%", "")
                        )
                    if "Number of reads mapped to multiple loci |" in line:
                        data["multimapped"] = int(
                            line.replace(
                                "Number of reads mapped to multiple loci |", ""
                            ).strip()
                        )
                    if "Number of splices: Non-canonical |" in line:
                        data["num_noncanonical_splices"] = int(
                            line.replace(
                                "Number of splices: Non-canonical |", ""
                            ).strip()
                        )
                    if "Number of reads unmapped: too short |" in line:
                        data["unmapped_tooshort"] = float(
                            line.replace("Number of reads unmapped: too short |", "")
                            .strip()
                            .replace("%", "")
                        )
                    if "% of reads mapped to too many loci |" in line:
                        data["multimapped_toomany_percent"] = float(
                            line.replace("% of reads mapped to too many loci |", "")
                            .strip()
                            .replace("%", "")
                        )
            return data

        data = []
        for sample in config["samples"]:
            d = get_numreads_percmapped(f"{config['outdir']}/{sample}/Log.final.out")
            data.append(d)
        df = pd.DataFrame(data, index=config["samples"])
        df.to_csv(f"{config['outdir']}/_output/readstat.tsv", sep="\t")

    tsc.ul.mkdir(f"{config['outdir']}/_output")
    if config["isrna"] == True:
        expression()
        genotype()
        readstats()
    else:
        genotype()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Farid Rashidi Mehrabadi All rights reserved.

# =========================================================================================
# Author     : Farid Rashidi Mehrabadi (farid.rashidimehrabadi@nih.gov)
# Last Update: Aug 14, 2020
# Description: cleaning
# =========================================================================================

import subprocess

from trisicell.ul._servers import *


def after03(config):
    # subprocess.getoutput(f'rm -rf {config['outdir']}/_annotating')
    subprocess.getoutput(f"rm -rf {config['outdir']}/_calling")
    subprocess.getoutput(f"rm -rf {config['outdir']}/_indexing")

    subprocess.getoutput(
        f"find {config['outdir']} -name \*Aligned.out.sam -exec rm -rf {{}} \;"
    )
    subprocess.getoutput(
        f"find {config['outdir']} -name \*Aligned.sortedByCoord.out.bam -exec rm -rf {{}} \;"
    )
    subprocess.getoutput(
        f"find {config['outdir']} -name \*Aligned.toTranscriptome.out.bam -exec rm -rf {{}} \;"
    )
    subprocess.getoutput(
        f"find {config['outdir']} -name \*dedupped.bam -exec rm -rf {{}} \;"
    )
    subprocess.getoutput(
        f"find {config['outdir']} -name \*expr.genes.results -exec rm -rf {{}} \;"
    )
    subprocess.getoutput(
        f"find {config['outdir']} -name \*expr.isoforms.results -exec rm -rf {{}} \;"
    )
    subprocess.getoutput(
        f"find {config['outdir']} -name \*expr.stat -exec rm -rf {{}} \;"
    )
    # subprocess.getoutput(f'find {config['outdir']} -name \*HaplotypeCaller.g.vcf -exec rm -rf {{}} \;')
    # subprocess.getoutput(f'find {config['outdir']} -name \*HaplotypeCaller.g.vcf.idx -exec rm -rf {{}} \;')
    subprocess.getoutput(
        f"find {config['outdir']} -name \*indel.intervals -exec rm -rf {{}} \;"
    )
    # subprocess.getoutput(f'find {config['outdir']} -name \*Log.final.out -exec rm -rf {{}} \;')
    subprocess.getoutput(
        f"find {config['outdir']} -name \*Log.out -exec rm -rf {{}} \;"
    )
    subprocess.getoutput(
        f"find {config['outdir']} -name \*Log.progress.out -exec rm -rf {{}} \;"
    )
    # subprocess.getoutput(f'find {config['outdir']} -name \*output.bai -exec rm -rf {{}} \;')
    # subprocess.getoutput(f'find {config['outdir']} -name \*output.bam -exec rm -rf {{}} \;')
    subprocess.getoutput(
        f"find {config['outdir']} -name \*output.metrics -exec rm -rf {{}} \;"
    )
    subprocess.getoutput(
        f"find {config['outdir']} -name \*ReadsPerGene.out.tab -exec rm -rf {{}} \;"
    )
    subprocess.getoutput(
        f"find {config['outdir']} -name \*realign.bai -exec rm -rf {{}} \;"
    )
    subprocess.getoutput(
        f"find {config['outdir']} -name \*realign.bam -exec rm -rf {{}} \;"
    )
    subprocess.getoutput(
        f"find {config['outdir']} -name \*recal.table -exec rm -rf {{}} \;"
    )
    subprocess.getoutput(
        f"find {config['outdir']} -name \*rg_added_sorted.bai -exec rm -rf {{}} \;"
    )
    subprocess.getoutput(
        f"find {config['outdir']} -name \*rg_added_sorted.bam -exec rm -rf {{}} \;"
    )
    subprocess.getoutput(
        f"find {config['outdir']} -name \*SJ.out.tab -exec rm -rf {{}} \;"
    )
    subprocess.getoutput(
        f"find {config['outdir']} -name \*split.bai -exec rm -rf {{}} \;"
    )
    subprocess.getoutput(
        f"find {config['outdir']} -name \*split.bam -exec rm -rf {{}} \;"
    )
    subprocess.getoutput(
        f"find {config['outdir']} -name \*_STARgenome -exec rm -rf {{}} \;"
    )

    if config["dotrimming"]:
        subprocess.getoutput(
            f"find {config['outdir']} -name \*_fastqc.html -exec rm -rf {{}} \;"
        )
        subprocess.getoutput(
            f"find {config['outdir']} -name \*_report.txt -exec rm -rf {{}} \;"
        )
        subprocess.getoutput(
            f"find {config['outdir']} -name \*_fastqc.zip -exec rm -rf {{}} \;"
        )
        subprocess.getoutput(
            f"find {config['outdir']} -name \*.fq.gz -exec rm -rf {{}} \;"
        )



import os

import pandas as pd
from tqdm import tqdm

DIFFERENT_LINEAGES = 0
ANCESTOR_DESCENDANT = 1
DESCENDANT_ANCESTOR = 2
SAME_NODE = 3
UNDEFINED_DEPENDENCY = 4


def read_CF_matrix_to_int_dictionary(path_CFmatrix_file):
    assert os.path.exists(path_CFmatrix_file), "ERROR. No file " + path_input_file

    input_file = open(path_CFmatrix_file, "r")
    mut_ids = input_file.readline().strip().split()[1:]
    m = len(mut_ids)
    assert len(set(mut_ids)) == m, "ERROR. Repeating mutations present in column names."

    E = {}
    for line in input_file.readlines():
        line_columns = line.strip().split()
        assert (
            len(line_columns) == m + 1
        ), "ERROR. Input file not in desired matrix format (note that tab or empty spaces are assumed as separators)"
        cell_id = line_columns[0]
        assert cell_id not in E, "ERROR. Repeating cell id " + cell_id
        E[cell_id] = {}

        for i in range(m):
            mut_id = mut_ids[i]
            value = int(line_columns[i + 1])
            assert value == 0 or value == 1, (
                "ERROR. Conflict free matrix "
                + path_CFmatrix_file
                + " contains non-binary entries"
            )
            E[cell_id][mut_id] = value

    input_file.close()

    return E, mut_ids


def get_dependency_from_conflict_free_matrix(mut1, mut2, matrix):
    """
    This function is used to compute phylogenetic dependency between two mutations based on the conflict-free matrix.

    Arguments:
    mut1: str
        ID of the first mutation. Should be among the columns of matrix.
    mut2: str
        ID of the second mutation. Should be among the columns of matrix.
    matrix: dictionary of dictionarie
        matrix is represented as dictionary of dictionaries with elements matrix[cell_id][mut_id]

    Returns:
    -------
    This function can return one of the following five values or report error (unknown case encountered).
    ANCESTOR_DESCENDANT: if there exists at least one cell where mut1=1 and mut2=0 and at least one cell where mut1=mut2=1
    DESCENDANT_ANCESTOR: if there exists at least one cell where mut1=0 and mut2=1 and at least one cell where mut1=mut2=1
    DIFFERENT_LINEAGES: if there exists at least one cell where mut1=1 and mut2=0 and at least one cell where mut1=0 and mut2=1
    SAME_NODE: if there exists at least one cell where mut1=mut2=1 and no cells where mut1=1 and mut2=0 or mut1=0 and mut2=1.
    UNDEFINED_DEPENDENCY: (ideally) if none of the above is returned.
    """

    pairs = [[0 for i in range(2)] for j in range(2)]
    for cell_id in list(matrix.keys()):
        a = int(matrix[cell_id][mut1])
        b = int(matrix[cell_id][mut2])
        if a not in [0, 1] or b not in [0, 1]:
            continue
        else:
            pairs[a][b] += 1

    if pairs[0][1] > 0 and pairs[1][0] > 0:
        assert pairs[1][1] == 0, (
            "ERROR. There exists conflict in the input matrix for mutations "
            + mut1
            + " and "
            + mut2
        )
        return DIFFERENT_LINEAGES

    if (
        pairs[0][1] > 0
    ):  # we know that in this case pairs[1][0] = 0 (see if statement above)
        if pairs[1][1] == 0:
            return UNDEFINED_DEPENDENCY
        else:
            return DESCENDANT_ANCESTOR

    if (
        pairs[1][0] > 0
    ):  # we know that in this case pairs[0][1]  = 0 (see previous to the above if statement above)
        if pairs[1][1] == 0:
            return UNDEFINED_DEPENDENCY
        else:
            return ANCESTOR_DESCENDANT

    if pairs[1][0] == 0 and pairs[0][1] == 0:
        if pairs[1][1] > 0:
            return SAME_NODE
        else:
            return UNDEFINED_DEPENDENCY

    assert (
        False
    ), "ERROR. Encountered case not covered in the implementation of function get_dependency."


def is_float(value):
    try:
        float(value)
        return True
    except ValueError:
        return False


def float_to_string(float_number, num_decimals=2):
    if is_float(float_number):
        return ("{:." + str(num_decimals) + "f}").format(float_number)
    else:
        return float_number


def prepare_dependencies(
    all_mut_ids, input_folder, output_file, max_num_submatrices, disable_tqdm
):
    assert len(set(all_mut_ids)) == len(
        all_mut_ids
    ), "ERROR. Mutation ids must be unique."

    dependencies = {}
    for mut1 in all_mut_ids:
        dependencies[mut1] = {}
        for mut2 in all_mut_ids:
            dependencies[mut1][mut2] = {}
            for dependency in [
                DIFFERENT_LINEAGES,
                ANCESTOR_DESCENDANT,
                DESCENDANT_ANCESTOR,
                SAME_NODE,
                UNDEFINED_DEPENDENCY,
            ]:
                dependencies[mut1][mut2][dependency] = 0

    CF_matrices_filenames = [
        x for x in os.listdir(input_folder) if x.endswith(".CFMatrix")
    ]
    # total = len(CF_matrices_filenames)
    # print(
    #     os.path.splitext(os.path.basename(noisy_SC_file))[0],
    #     len(CF_matrices_filenames),
    #     max_num_submatrices,
    # )

    completed = 0
    for filename in tqdm(
        CF_matrices_filenames,
        ascii=True,
        ncols=100,
        desc="DEPENDENCIES",
        total=max_num_submatrices,
        position=0,
        disable=disable_tqdm,
    ):
        if max_num_submatrices is not None and completed == max_num_submatrices:
            break

        # print("Completed " + str(completed) + " out of " + str(total))
        try:
            matrix, mut_ids = read_CF_matrix_to_int_dictionary(
                os.path.join(input_folder, filename)
            )
        except:
            completed -= 1

        for i in range(len(mut_ids)):
            mut1 = mut_ids[i]
            dependencies[mut1][mut1][SAME_NODE] += 1

            for j in range(i + 1, len(mut_ids)):
                mut2 = mut_ids[j]
                dependency = get_dependency_from_conflict_free_matrix(
                    mut1, mut2, matrix
                )

                if (
                    dependency == SAME_NODE
                    or dependency == DIFFERENT_LINEAGES
                    or dependency == UNDEFINED_DEPENDENCY
                ):
                    dependencies[mut1][mut2][dependency] += 1
                    dependencies[mut2][mut1][dependency] += 1
                elif dependency == ANCESTOR_DESCENDANT:
                    dependencies[mut1][mut2][ANCESTOR_DESCENDANT] += 1
                    dependencies[mut2][mut1][DESCENDANT_ANCESTOR] += 1
                elif dependency == DESCENDANT_ANCESTOR:
                    dependencies[mut1][mut2][DESCENDANT_ANCESTOR] += 1
                    dependencies[mut2][mut1][ANCESTOR_DESCENDANT] += 1
                else:
                    assert False, (
                        "ERROR. Encountered unknown dependency in file "
                        + os.path.join(input_folder, filename)
                        + " for mutations "
                        + mut1
                        + " and "
                        + mut2
                    )
        completed += 1

    output_file = open(output_file, "w")
    output_file.write(
        "\t".join(
            [
                "MUT1",
                "MUT2",
                "DIFFERENT_LINEAGES",
                "ANCESTOR_DESCENDANT",
                "DESCENDANT_ANCESTOR",
                "SAME_NODE",
                "UNDEFINED_DEPENDENCY",
                "TOTAL",
                "DOMINANT_DEPENDENCY",
                "PERCENTAGE_DOMINANT",
            ]
        )
    )
    output_file.write("\n")

    for mut1 in all_mut_ids:
        for mut2 in all_mut_ids:
            line_to_write = mut1 + "\t" + mut2

            dominant_dependency = None
            max_count = 0
            total_count = 0
            for dependency in [
                DIFFERENT_LINEAGES,
                ANCESTOR_DESCENDANT,
                DESCENDANT_ANCESTOR,
                SAME_NODE,
                UNDEFINED_DEPENDENCY,
            ]:
                current_count = dependencies[mut1][mut2][dependency]
                line_to_write += "\t" + str(current_count)

                if current_count > max_count:
                    dominant_dependency = dependency
                    max_count = current_count

                total_count += current_count

            line_to_write += "\t" + str(total_count)

            if total_count == 0:
                line_to_write += "\t" + "UNDEFINED"
                line_to_write += "\t" + "UNDEFINED"
                if mut1 < mut2:
                    # print("WARNING. For file " + noisy_SC_file + " mutations " + mut1 + " and " + mut2 + " were never subsampled together")
                    pass
            else:
                line_to_write += "\t" + str(dominant_dependency)
                line_to_write += "\t" + float_to_string(
                    (100.0 * max_count) / total_count
                )

            output_file.write(line_to_write + "\n")

    output_file.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Farid Rashidi Mehrabadi All rights reserved.

# =========================================================================================
# Author     : Farid Rashidi Mehrabadi (farid.rashidimehrabadi@nih.gov)
# Last Update: May 17, 2020
# Description: calling mutations as a joint cohort by GATK GenotypeGVCFs for each chr
# =========================================================================================

from trisicell.ul._servers import *


def run06(config, afterok):
    def cmds(chrom):
        cmds = ""
        cmds += cmd([f"mkdir -p {config['outdir']}/_calling"])
        if len(config["normals"]) > 0:
            files = " -V ".join(
                [
                    f"{config['outdir']}/{s}/MuTect2.g.vcf"
                    for s in config["samples"]
                    if s not in config["normals"]
                ]
            )
        else:
            files = " -V ".join(
                [
                    f"{config['outdir']}/{s}/HaplotypeCaller.g.vcf"
                    for s in config["samples"]
                ]
            )
        cmds += cmd(
            [
                f"{config['java06']} {config['GATK']}",
                f"-T GenotypeGVCFs",
                f"-R {config['ref']}",
                f"-V {files}",
                f"-o {config['outdir']}/_calling/jointcalls.{chrom}.g.vcf",
                f"-nt {config['threads06']}",
                f"-L {chrom}",
                f"--disable_auto_index_creation_and_locking_when_reading_rods",
            ]
        )
        cmds += cmd([f"echo Done!"], islast=True)
        return cmds

    df = pd.DataFrame()
    df["chrom"] = config["chroms"]
    df["cmd"] = df.apply(lambda x: cmds(x["chrom"]), axis=1)

    jobname = os.path.basename(__file__)[:-3]
    cmdmain = write_cmds_get_main(
        df,
        jobname,
        config["time06"],
        config["mem06"],
        "python/3.7",
        config["threads06"],
        config["email"],
        config["tmpdir"],
        afterok,
    )
    return cmdmain



import pandas as pd

import trisicell as tsc


def example(kind):
    """A toy example for sanity checking and playing with Trisicell.

    Parameters
    ----------
    kind : :obj:`str`
        The options are:

            - `'genotype'`: This option will return information about genotype matrix.
            - `'expression'`: This option will return information about expression matrix.

    Returns
    -------
    :class:`anndata.AnnData`
        An object that cells are in `.obs` and mutations are in `.var`.

    Raises
    ------
    ValueError
        If `kind` is not `genotype` or `expression`.
    """

    if kind.lower() == "genotype":
        return tsc.io.read(tsc.ul.get_file("trisicell.datasets/data/genotype.h5ad.gz"))
    elif kind.lower() == "expression":
        return tsc.io.read(
            tsc.ul.get_file("trisicell.datasets/data/expression.h5ad.gz")
        )
    else:
        raise ValueError("kind must be either genotype or expression")


def test():
    df = tsc.io.read(tsc.ul.get_file("trisicell.datasets/test/test.SC"))
    return df


def melanoma20():
    """Mouse Melanoma dataset with 20 sublines.

    This dataset was introduced in :cite:`Wolf_2019` and was used in:

    * :cite:`PhISCS-BnB` Figure 1.

    The size is n_cells × n_muts = 20 × 2367

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.

        `.layers['solution_fig1']` is the solution presented in Figure 1 of PhISCS-BnB paper.
    """

    adata = tsc.io.read(tsc.ul.get_file("trisicell.datasets/real/melanoma20.h5ad"))
    return adata


def colorectal1():
    """Human Colorectal Cancer dataset patient 1.

    This dataset was introduced in :cite:`Leung_2017` and was used in:

    * :cite:`B-SCITE` Figure 8a.

    The size is n_cells × n_muts = 72 × 12

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.
    """

    # TODO: extract.
    pass


def colorectal2():
    """Human Colorectal Cancer dataset patient 2.

    This dataset was introduced in :cite:`Leung_2017` and was used in:

    * :cite:`PhISCS` Figure 7.
    * :cite:`B-SCITE` Figure 8b.

    The size is n_cells × n_muts = 78 × 25

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.

            - `.layers['solution_fig7a']` is the solution presented in Figure 7a of PhISCS paper.
            - `.layers['solution_fig7b']` is the solution presented in Figure 7b of PhISCS paper.
            - `.uns['params_fig7a']` is parameters used as input to get 'solution_7a'.
            - `.uns['params_fig7b']` is parameters used as input to get 'solution_7b'.
            - `.var` includes information of the bulk samples.
    """

    adata = tsc.io.read(tsc.ul.get_file("trisicell.datasets/real/colorectal2.h5ad"))
    # TODO: 86 x 25 in B-SCITE?
    return adata


def acute_lymphocytic_leukemia1():
    """Human Acute Lymphocytic Leukemia dataset.

    This dataset was introduced in :cite:`Gawad_2014` and was used in:

    * :cite:`B-SCITE` Figure 5.

    The size is n_cells × n_muts = 111 × 20

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.
    """

    # TODO: extract.
    pass


def acute_lymphocytic_leukemia2():
    """Human Acute Lymphocytic Leukemia dataset.

    This dataset was introduced in :cite:`Gawad_2014` and was used in:

    * :cite:`PhISCS` Figure 9.
    * :cite:`B-SCITE` in Figure 6.

    The size is n_cells × n_muts = 102 × 16

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.

            - `.layers['solution_fig9']` is the solution presented in Figure 9 of PhISCS paper.
            - `.uns['params_fig9']` is parameters used as input to get 'solution_fig9'.
            - `.var` includes information of the bulk samples.
    """

    adata = tsc.io.read(
        tsc.ul.get_file("trisicell.datasets/real/acute_lymphocytic_leukemia2.h5ad")
    )
    # TODO: 115 x 16 in B-SCITE?
    return adata


def acute_lymphocytic_leukemia3():
    # TODO: extract.
    pass


def acute_lymphocytic_leukemia4():
    # TODO: extract.
    pass


def acute_lymphocytic_leukemia5():
    # TODO: extract.
    pass


def acute_lymphocytic_leukemia6():
    # TODO: extract.
    pass


def myeloproliferative_neoplasms18():
    """JAK2-Negative Myeloproliferative Neoplasm.

    This dataset was introduced in :cite:`Hou_2012` and was used in:

    * :cite:`PhISCS` Figure 9.

    The size is n_cells × n_muts = 58 × 18

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.

            - `.layers['solution_fig9']` is the solution presented in Figure 9 of PhISCS paper.
            - `.uns['params_fig9']` is parameters used as input to get 'solution_fig9'.
            - `.var` includes information of the bulk samples.
    """

    pass


def triple_negative_breast_cancer():
    """Triple-negative Breast Cancer.

    This dataset was introduced in []_ and was used in:
    * :cite:`B-SCITE` Figure 7.

    The size is n_cells × n_muts = 16 × 18

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy.
    """

    # TODO: extract.
    pass


def high_grade_serous_ovarian_cancer():
    """Triple-negative Breast Cancer.

    This dataset was introduced in :cite:`Laks_2019`
    and preprocessed in :cite:`McPherson_2019`.
    The phylogeny is presented in Figure 3H.

    The size is n_cells × n_muts = 891 × 13666

    Note that 402 mutations were deleted during evolution
    in the inferred tree by original study So, here they were filtered out
    (meaning they are 14068 mutations in total in the original study).

    Returns
    -------
    :class:`anndata.AnnData`
        An anndata in which `.X` is the input noisy
        (it is obtained based on the number of mutant/total reads).

            - `.obs['clone_id']` is the clone id to which the cell is assigned in Figure 3H.
            - `.obs['group_color']` is unique colors for each 'clone_id'.
            - `.obs['cell_name']` is a new name for each cell based on the 'group_color'.
            - `.layers['mutant']` is the number of mutant reads at each locus in each cell.
            - `.layers['total']` is the total number of reads at each locus in each cell.
            - `.layers['ground']` is the solution inferred in Figure 3H of the original paper.
            - `.uns['params_ground']` is parameters inferred by comparing ground and noisy matrices.
            - `.var` includes information of the bulk samples.

    Examples
    --------
    >>> sc = tsc.datasets.high_grade_serous_ovarian_cancer()
    >>> print(sc)
    AnnData object with n_obs × n_vars = 891 × 13666
        obs: 'clone_id', 'group_color', 'cell_name'
        uns: 'params_ground'
        layers: 'ground', 'mutant', 'total'
    """

    adata = tsc.io.read(tsc.ul.get_file("trisicell.datasets/real/ovarian.h5ad.gz"))
    return adata



import os

import pandas as pd

import trisicell as tsc
from trisicell.tl.solver.booster._dependencies import prepare_dependencies
from trisicell.tl.solver.booster._subsamples import subsampling


def booster(
    df_input,
    alpha,
    beta,
    solver="PhISCS",
    sample_on="muts",
    sample_size=10,
    n_samples=10,
    begin_sample=0,
    n_jobs=10,
    time_out=120,
    save_inter=True,
    dir_inter=".",
    base_inter=None,
    disable_tqdm=False,
    weight=50,
    no_subsampling=False,
    no_dependencies=False,
):
    """Divide and Conquer Booster solver.

    Parameters
    ----------
    df_input : :class:`pandas.DataFrame`
        input noisy dataframe
    alpha : float
        false positive rate
    beta : float
        false negative rate
    solver : :obj:`str`, optional
        [description], by default "PhISCS"
    sample_size : :obj:`int`, optional
        [description], by default 10
    n_samples : :obj:`int`, optional
        [description], by default 10
    begin_sample : :obj:`int`, optional
        [description], by default 0
    n_jobs : :obj:`int`, optional
        [description], by default 10
    time_out : :obj:`int`, optional
        [description], by default 120
    save_inter : :obj:`bool`, optional
        [description], by default True
    dir_inter : :obj:`str`, optional
        [description], by default "."
    base_inter : :obj:`str`, optional
        [description], by default None
    disable_tqdm : :obj:`bool`, optional
        [description], by default False
    weight : :obj:`int`, optional
        [description], by default 50

    Returns
    -------
    :class:`pandas.DataFrame`
        [description]


    See Also
    --------
    :func:`trisicell.tl.scite`.
    """

    if not base_inter:
        tmpdir = tsc.ul.tmpdir(
            prefix="trisicell.", suffix=".booster", dirname=dir_inter
        )
    else:
        tmpdir = tsc.ul.mkdir(os.path.join(dir_inter, base_inter))

    #### subsampling matrices and solving them
    if not no_subsampling:
        subsampling(
            df_input,
            alpha=alpha,
            beta=beta,
            solver=solver,
            sample_on=sample_on,
            sample_size=sample_size,
            n_samples=n_samples,
            begin_sample=begin_sample,
            n_jobs=n_jobs,
            time_out=time_out,
            tmpdir=tmpdir,
            disable_tqdm=disable_tqdm,
        )

    #### preparing dependencies file
    if not no_dependencies:
        n_muts = df_input.shape[1]
        max_num_submatrices = int(weight * (n_muts ** 2) / (sample_size ** 2))
        prepare_dependencies(
            df_input.columns,
            tmpdir,
            f"{tmpdir}/_booster.dependencies",
            max_num_submatrices,
            disable_tqdm,
        )

    #### building the final cfmatrix

    df_output = pd.DataFrame(df_input.values)
    df_output.columns = df_input.columns
    df_output.index = df_input.index
    df_output.index.name = "cellIDxmutID"

    if not save_inter:
        tsc.ul.cleanup(tmpdir)

    # tsc.ul.stat(df_input, df_output, alpha, beta, running_time)

    return df_output



import numpy as np
import pandas as pd
import seaborn as sns
from scipy.cluster.hierarchy import cut_tree, dendrogram, fcluster, linkage


def calc_hclust(dist, names, method="ward"):
    clust = linkage(dist[np.triu_indices(dist.shape[0], 1)], method=method)
    clusters = {}

    for i in range(2, dist.shape[0]):
        fc = fcluster(clust, i, criterion="maxclust")
        clusters[i] = pd.Series(fc, index=names)

    return clusters

    # sns.set(style='ticks', font='Helvetica', font_scale=0.4)
    # dend = dendrogram(clust, color_threshold=threshold, labels=names)

    # dend = dendrogram(clust)
    # idx = [int(x) for x in dend["ivl"]]
    # corrcoef = 1 - dist
    # sns.clustermap(
    #     corrcoef[idx, :][:, idx],
    #     vmin=vmin,
    #     vmax=vmax,
    #     cmap="RdYlBu_r",
    #     row_cluster=False,
    #     col_cluster=False,
    #     cbar_pos=(1, 0.03, 0.05, 0.94),
    #     figsize=(5, 5),
    #     xticklabels=names[idx],
    #     yticklabels=False,
    #     dendrogram_ratio=0.000001,
    # )



import datetime
from platform import python_version

from trisicell import settings

_VERBOSITY_LEVELS_FROM_STRINGS = {
    "error": 0,
    "warn": 1,
    "info": 2,
    "hint": 3,
    "debug": 4,
}


def info(*args, **kwargs):
    return msg(*args, v="info", **kwargs)


def error(*args, **kwargs):
    args = ("Error:",) + args
    return msg(*args, v="error", **kwargs)


def warn(*args, **kwargs):
    args = ("WARNING:",) + args
    return msg(*args, v="warn", **kwargs)


def hint(*args, **kwargs):
    return msg(*args, v="hint", **kwargs)


def debug(*args, **kwargs):
    args = ("DEBUG:",) + args
    return msg(*args, v="debug", **kwargs)


def msg(*msg, v, time=False, end="\n"):
    if v is None:
        v = 4
    if isinstance(v, str):
        v = _VERBOSITY_LEVELS_FROM_STRINGS[v]
    if v == 3:
        msg = ("-->",) + msg
    if v >= 4:
        msg = ("   ",) + msg
    if _settings_verbosity_greater_or_equal_than(v):
        if len(msg) > 0:
            if time:
                msg = (f"[{datetime.datetime.now().strftime('%m/%d %H:%M:%S')}]",) + msg
            _write_log(*msg, end=end)


def _write_log(*msg, end="\n"):
    from trisicell.settings import logfile

    if logfile == "":
        print(*msg, end=end, flush=True)
    else:
        out = ""
        for s in msg:
            out += f"{s} "
        with open(logfile, "a") as f:
            f.write(out + end)
            f.flush()


def _settings_verbosity_greater_or_equal_than(v):
    from trisicell.settings import verbosity

    if isinstance(verbosity, str):
        settings_v = _VERBOSITY_LEVELS_FROM_STRINGS[verbosity]
    else:
        settings_v = verbosity
    return settings_v >= v


def _check_if_latest_version():
    from trisicell import __version__

    latest_version = timeout(
        get_latest_pypi_version, timeout_duration=2, default="0.0.0"
    )
    if __version__.rsplit(".dev")[0] < latest_version.rsplit(".dev")[0]:
        warn(
            "There is a newer trisicell version available on PyPI:\n",
            "Your version: \t\t",
            __version__,
            "\nLatest version: \t",
            latest_version,
        )


def _get_date_string():
    return datetime.datetime.now().strftime("%Y-%m-%d %H:%M")


def print_version():
    from trisicell import __version__

    _write_log(
        f"Running trisicell {__version__} "
        f"(python {python_version()}) on {_get_date_string()}.",
    )
    # _check_if_latest_version()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------
# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
from datetime import datetime
from glob import glob
from shutil import copyfile

from pybtex.plugin import register_plugin
from pybtex.style.formatting.unsrt import Style as UnsrtStyle
from pybtex.style.labels import BaseLabelStyle

import trisicell

# -- Retrieve notebooks ------------------------------------------------
# for nb in glob(os.path.join(".", "*.ipynb")):
#     os.remove(nb)
# notebooks_url = "../../trisicell_notebooks/"
# notebooks = [
#     "Mel.Subline.ipynb",
#     "Mel.Treated.ipynb",
# ]
# for nb in notebooks:
#     copyfile(os.path.join(notebooks_url, nb), os.path.join(".", nb))


# -- Project information -----------------------------------------------------
nitpicky = True  # Warn about broken links. This is here for a reason: Do not change.
needs_sphinx = "2.0"  # Nicer param docs
suppress_warnings = ["ref.citation"]
project = "Trisicell"
author = "National Cancer Institute"  # trisicell.__author__
title = "Scalable tumor phylogeny reconstruction and validation from single-cell data"
copyright = f"{datetime.now():%Y}, {author}"
release = "master"
version = f"master ({trisicell.__version__})"


# -- General configuration ---------------------------------------------------
# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named "sphinx.ext.*") or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.doctest",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.autosummary",
    "sphinx.ext.githubpages",
    "sphinx_autodoc_typehints",
    "nbsphinx",
    "sphinx_paramlinks",
    "sphinx_copybutton",
    "sphinx_click.ext",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinxcontrib.bibtex",
    # "sphinx_gallery.gen_gallery",
    # "sphinx_last_updated_by_git",
]

# Generate the API documentation when building
autosummary_generate = True
autodoc_member_order = "bysource"
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_use_rtype = True
napoleon_use_param = True
napoleon_custom_sections = [("Params", "Parameters")]
todo_include_todos = False

# bibliography
bibtex_bibfiles = ["references.bib"]
bibtex_default_style = "mystyle"

# spelling
spelling_lang = "en_US"
spelling_warning = True
spelling_add_pypi_package_names = True
spelling_show_suggestions = True
spelling_exclude_patterns = ["references.rst"]
# see: https://pyenchant.github.io/pyenchant/api/enchant.tokenize.html
spelling_filters = [
    "enchant.tokenize.URLFilter",
    "enchant.tokenize.EmailFilter",
    "docs.source.utils.ModnameFilter",
    "docs.source.utils.SignatureFilter",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]
source_suffix = [".rst", ".ipynb"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
master_doc = "index"
default_role = "literal"
pygments_style = "sphinx"
todo_include_todos = False

intersphinx_mapping = dict(
    python=("https://docs.python.org/3", None),
    numpy=("https://docs.scipy.org/doc/numpy/", None),
    scipy=("https://docs.scipy.org/doc/scipy/reference/", None),
    networkx=("https://networkx.github.io/documentation/stable/", None),
    pandas=("https://pandas.pydata.org/pandas-docs/stable/", None),
    matplotlib=("https://matplotlib.org/", None),
    sklearn=("https://scikit-learn.org/stable/", None),
    seaborn=("https://seaborn.pydata.org/", None),
    anndata=("https://anndata.readthedocs.io/en/latest/", None),
)


# -- Options for HTML output -------------------------------------------------
html_theme = "sphinx_rtd_theme"
github_repo = "trisicell"
github_nb_repo = "trisicell_notebooks"
html_logo = "_static/images/logo.svg"
html_theme_options = {
    "logo_only": True,
    "display_version": True,
    "navigation_depth": 2,
}
html_context = {
    "display_github": True,
    "github_user": "faridrashidi",
    "github_repo": "trisicell",
    "github_version": "master",
    "conf_py_path": "/docs/",
}
html_show_sphinx = False
html_static_path = ["_static"]
html_css_files = ["css/custom.css"]


# -- Options for other output ------------------------------------------
htmlhelp_basename = f"{project}doc"
title_doc = f"{project} documentation"

latex_documents = [(master_doc, f"{project}.tex", title_doc, author, "manual")]
man_pages = [(master_doc, project, title_doc, [author], 1)]
texinfo_documents = [
    (master_doc, project, title_doc, author, project, title, "Miscellaneous")
]


# a simple label style which uses the bibtex keys for labels
class MyLabelStyle(BaseLabelStyle):
    def format_labels(self, sorted_entries):
        for entry in sorted_entries:
            yield entry.key


class MyStyle(UnsrtStyle):

    default_label_style = MyLabelStyle


register_plugin("pybtex.style.formatting", "mystyle", MyStyle)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Farid Rashidi Mehrabadi All rights reserved.

# =========================================================================================
# Author     : Farid Rashidi Mehrabadi (farid.rashidimehrabadi@nih.gov)
# Last Update: Aug 03, 2020
# Description: merging joint mutations called
# =========================================================================================

from trisicell.ul._servers import *


def run07(config, afterok):
    def cmds():
        cmds = ""
        cmds += cmd([f"module load bcftools/1.9"])
        files = " ".join(
            [
                f"{config['outdir']}/_calling/jointcalls.{chrom}.g.vcf"
                for chrom in config["chroms"]
            ]
        )
        cmds += cmd(
            [
                f"bcftools concat",
                f"-o {config['outdir']}/_calling/jointcalls.g.vcf",
                f"{files}",
            ]
        )
        cmds += cmd([f"echo Done!"], islast=True)
        return cmds

    df = pd.DataFrame()
    df["cmd"] = [cmds()]

    jobname = os.path.basename(__file__)[:-3]
    cmdmain = write_cmds_get_main(
        df,
        jobname,
        config["time07"],
        config["mem07"],
        "",
        1,
        config["email"],
        config["tmpdir"],
        afterok,
    )
    return cmdmain



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Farid Rashidi Mehrabadi All rights reserved.

# =========================================================================================
# Author     : Farid Rashidi Mehrabadi (farid.rashidimehrabadi@nih.gov)
# Last Update: July 11, 2020
# Description: mapping for the second pass of the STAR + GATK best practices
# =========================================================================================

from trisicell.ul._servers import *


def run04(config, afterok):
    def cmds(sample):
        cmds = ""
        cmds += cmd([f"mkdir -p {config['outdir']}/{sample}"])
        if config["isrna"] == True:
            if config["dotrimming"] == True:
                if config["pairedend"] == True:
                    cmds += cmd(
                        [
                            f"STAR",
                            f"--runMode alignReads",
                            f"--genomeDir {config['outdir']}/_indexing/2",
                            f"--readFilesCommand zcat",
                            f"--readFilesIn {config['outdir']}/{sample}/{config['infqpre1']}{sample}{config['infqpost3']} {config['outdir']}/{sample}/{config['infqpre2']}{sample}{config['infqpost4']}",
                            f"--outFileNamePrefix {config['outdir']}/{sample}/",
                            f"--limitBAMsortRAM 30000000000",
                            f"--outSAMtype BAM SortedByCoordinate",
                            f"--sjdbGTFfile {config['annot']}",
                            f"--outFilterMultimapNmax 1",
                            f"--outSAMunmapped None",
                            f"--quantMode TranscriptomeSAM GeneCounts",
                            f"--runThreadN 1",
                            f"--sjdbOverhang {config['readlength']}",
                        ]
                    )
                    cmds += cmd(
                        [
                            f"rm -rf",
                            f"{config['outdir']}/{sample}/{config['infqpre1']}{sample}{config['infqpost3']}",
                            f"{config['outdir']}/{sample}/{config['infqpre2']}{sample}{config['infqpost4']}",
                        ]
                    )
                else:
                    cmds += cmd(
                        [
                            f"STAR",
                            f"--runMode alignReads",
                            f"--genomeDir {config['outdir']}/_indexing/2",
                            f"--readFilesCommand zcat",
                            f"--readFilesIn {config['outdir']}/{sample}/{config['infqpre']}{sample}{config['infqpost5']}",
                            f"--outFileNamePrefix {config['outdir']}/{sample}/",
                            f"--limitBAMsortRAM 30000000000",
                            f"--outSAMtype BAM SortedByCoordinate",
                            f"--sjdbGTFfile {config['annot']}",
                            f"--outFilterMultimapNmax 1",
                            f"--outSAMunmapped None",
                            f"--quantMode TranscriptomeSAM GeneCounts",
                            f"--runThreadN 1",
                            f"--sjdbOverhang {config['readlength']}",
                        ]
                    )
                    cmds += cmd(
                        [
                            f"rm -rf",
                            f"{config['outdir']}/{sample}/{config['infqpre']}{sample}{config['infqpost5']}",
                        ]
                    )
            else:
                if config["pairedend"] == True:
                    cmds += cmd(
                        [
                            f"STAR",
                            f"--runMode alignReads",
                            f"--genomeDir {config['outdir']}/_indexing/2",
                            f"--readFilesCommand zcat",
                            f"--readFilesIn {config['infq']}/{config['infqpre1']}{sample}{config['infqpost1']} {config['infq']}/{config['infqpre2']}{sample}{config['infqpost2']}",
                            f"--outFileNamePrefix {config['outdir']}/{sample}/",
                            f"--limitBAMsortRAM 30000000000",
                            f"--outSAMtype BAM SortedByCoordinate",
                            f"--sjdbGTFfile {config['annot']}",
                            f"--outFilterMultimapNmax 1",
                            f"--outSAMunmapped None",
                            f"--quantMode TranscriptomeSAM GeneCounts",
                            f"--runThreadN 1",
                            f"--sjdbOverhang {config['readlength']}",
                        ]
                    )
                else:
                    cmds += cmd(
                        [
                            f"STAR",
                            f"--runMode alignReads",
                            f"--genomeDir {config['outdir']}/_indexing/2",
                            f"--readFilesCommand zcat",
                            f"--readFilesIn {config['infq']}/{config['infqpre']}{sample}{config['infqpost']}",
                            f"--outFileNamePrefix {config['outdir']}/{sample}/",
                            f"--limitBAMsortRAM 30000000000",
                            f"--outSAMtype BAM SortedByCoordinate",
                            f"--sjdbGTFfile {config['annot']}",
                            f"--outFilterMultimapNmax 1",
                            f"--outSAMunmapped None",
                            f"--quantMode TranscriptomeSAM GeneCounts",
                            f"--runThreadN 1",
                            f"--sjdbOverhang {config['readlength']}",
                        ]
                    )
        cmds += cmd(
            [
                f"rm -rf",
                f"{config['outdir']}/{sample}/Aligned.out.sam",
            ]
        )

        cmds += cmd(
            [
                f"{config['java04']} {config['PICARD']}",
                f"AddOrReplaceReadGroups",
                f"INPUT={config['outdir']}/{sample}/Aligned.sortedByCoord.out.bam",
                f"OUTPUT={config['outdir']}/{sample}/dedupped.bam",
                f"SORT_ORDER=coordinate",
                f"RGID={sample}",
                f"RGLB=trancriptome" if config["isrna"] == True else f"RGLB=genome",
                f"RGPL=ILLUMINA",
                f"RGPU=machine",
                f"RGSM={sample}",
            ]
        )

        cmds += cmd(
            [
                f"{config['java04']} {config['PICARD']}",
                f"MarkDuplicates",
                f"INPUT={config['outdir']}/{sample}/dedupped.bam",
                f"OUTPUT={config['outdir']}/{sample}/rg_added_sorted.bam",
                f"METRICS_FILE={config['outdir']}/{sample}/output.metrics",
                f"VALIDATION_STRINGENCY=SILENT",
                f"CREATE_INDEX=true",
            ]
        )
        cmds += cmd(
            [
                f"rm -rf",
                f"{config['outdir']}/{sample}/dedupped.bam",
            ]
        )

        if config["isrna"] == True:
            cmds += cmd(
                [
                    f"{config['java04']} {config['GATK']}",
                    f"-T SplitNCigarReads",
                    f"-R {config['ref']}",
                    f"-I {config['outdir']}/{sample}/rg_added_sorted.bam",
                    f"-o {config['outdir']}/{sample}/split.bam",
                    f"-rf ReassignOneMappingQuality",
                    f"-RMQF 255",
                    f"-RMQT 60",
                    f"-U ALLOW_N_CIGAR_READS",
                ]
            )

        if config["buildver"] == "hg19":
            cmds += cmd(
                [
                    f"{config['java04']} {config['GATK']}",
                    f"-T RealignerTargetCreator",
                    f"-R {config['ref']}",
                    f"-I {config['outdir']}/{sample}/split.bam"
                    if config["isrna"] == True
                    else f"-I {config['outdir']}/{sample}/rg_added_sorted.bam",
                    f"-o {config['outdir']}/{sample}/indel.intervals",
                    f"-known {config['hgKGIndels']}",
                    f"-known {config['hgMillsIndels']}",
                    f"-U ALLOW_SEQ_DICT_INCOMPATIBILITY",
                ]
            )

            cmds += cmd(
                [
                    f"{config['java04']} {config['GATK']}",
                    f"-T IndelRealigner",
                    f"-R {config['ref']}",
                    f"-I {config['outdir']}/{sample}/split.bam"
                    if config["isrna"] == True
                    else f"-I {config['outdir']}/{sample}/rg_added_sorted.bam",
                    f"-o {config['outdir']}/{sample}/realign.bam",
                    f"-targetIntervals {config['outdir']}/{sample}/indel.intervals",
                    f"-known {config['hgKGIndels']}",
                    f"-known {config['hgMillsIndels']}",
                ]
            )

            cmds += cmd(
                [
                    f"{config['java04']} {config['GATK']}",
                    f"-T BaseRecalibrator",
                    f"-R {config['ref']}",
                    f"-I {config['outdir']}/{sample}/realign.bam",
                    f"-o {config['outdir']}/{sample}/recal.table",
                    f"-knownSites {config['hgDBSNPs138']}",
                    f"-knownSites {config['hgKGIndels']}",
                    f"-knownSites {config['hgMillsIndels']}",
                ]
            )
        else:
            cmds += cmd(
                [
                    f"{config['java04']} {config['GATK']}",
                    f"-T RealignerTargetCreator",
                    f"-R {config['ref']}",
                    f"-I {config['outdir']}/{sample}/split.bam"
                    if config["isrna"] == True
                    else f"-I {config['outdir']}/{sample}/rg_added_sorted.bam",
                    f"-o {config['outdir']}/{sample}/indel.intervals",
                    f"-known {config['mmDBIndels142']}",
                    f"-U ALLOW_SEQ_DICT_INCOMPATIBILITY",
                ]
            )

            cmds += cmd(
                [
                    f"{config['java04']} {config['GATK']}",
                    f"-T IndelRealigner",
                    f"-R {config['ref']}",
                    f"-I {config['outdir']}/{sample}/split.bam"
                    if config["isrna"] == True
                    else f"-I {config['outdir']}/{sample}/rg_added_sorted.bam",
                    f"-o {config['outdir']}/{sample}/realign.bam",
                    f"-targetIntervals {config['outdir']}/{sample}/indel.intervals",
                    f"-known {config['mmDBIndels142']}",
                ]
            )

            cmds += cmd(
                [
                    f"{config['java04']} {config['GATK']}",
                    f"-T BaseRecalibrator",
                    f"-R {config['ref']}",
                    f"-I {config['outdir']}/{sample}/realign.bam",
                    f"-o {config['outdir']}/{sample}/recal.table",
                    f"-knownSites {config['mmDBSNPs142']}",
                    f"-knownSites {config['mmDBIndels142']}",
                ]
            )
        cmds += cmd(
            [
                f"rm -rf",
                f"{config['outdir']}/{sample}/rg_added_sorted.bam",
                f"{config['outdir']}/{sample}/rg_added_sorted.bai",
                f"{config['outdir']}/{sample}/split.bam",
                f"{config['outdir']}/{sample}/split.bai",
            ]
        )
        cmds += cmd(
            [
                f"{config['java04']} {config['GATK']}",
                f"-T PrintReads",
                f"-R {config['ref']}",
                f"-I {config['outdir']}/{sample}/realign.bam",
                f"-o {config['outdir']}/{sample}/output.bam",
                f"-BQSR {config['outdir']}/{sample}/recal.table",
            ]
        )
        cmds += cmd(
            [
                f"rm -rf",
                f"{config['outdir']}/{sample}/realign.bam",
                f"{config['outdir']}/{sample}/realign.bai",
                f"{config['outdir']}/{sample}/Log.progress.out",
                f"{config['outdir']}/{sample}/Log.out",
                f"{config['outdir']}/{sample}/indel.intervals",
                f"{config['outdir']}/{sample}/output.metrics",
                f"{config['outdir']}/{sample}/ReadsPerGene.out.tab",
                f"{config['outdir']}/{sample}/recal.table",
                f"{config['outdir']}/{sample}/SJ.out.tab",
                f"{config['outdir']}/{sample}/_STARgenome",
                # f'{config['outdir']}/{sample}/Aligned.sortedByCoord.out.bam',
            ]
        )
        cmds += cmd(
            [
                f"mv",
                f"{config['outdir']}/{sample}/Aligned.sortedByCoord.out.bam",
                f"{config['outdir']}/{sample}/{sample}.align.bam",
            ]
        )
        cmds += cmd(
            [
                f"samtools index",
                f"{config['outdir']}/{sample}/{sample}.align.bam",
            ]
        )
        cmds += cmd([f"echo Done!"], islast=True)
        return cmds

    df = pd.DataFrame()
    df["sample"] = config["samples"]
    df["cmd"] = df.apply(lambda x: cmds(x["sample"]), axis=1)

    jobname = os.path.basename(__file__)[:-3]
    cmdmain = write_cmds_get_main(
        df,
        jobname,
        config["time04"],
        config["mem04"],
        "STAR/2.7.3a,samtools/1.11",
        1,
        config["email"],
        config["tmpdir"],
        afterok,
    )
    return cmdmain



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Farid Rashidi Mehrabadi All rights reserved.

# =========================================================================================
# Author     : Farid Rashidi Mehrabadi (farid.rashidimehrabadi@nih.gov)
# Last Update: Aug 14, 2020
# Description: cleaning
# =========================================================================================

import glob
import subprocess

from trisicell.ul._servers import *


def _is_ok(name):
    file = open(name, "r")
    body = file.read()
    file.close()
    a = body.count("&& echo Done! )")
    b = body.count("Done!\n")
    if a == 0 and b == 1:
        return True
    else:
        return a == b


def after01(config):
    if config["isrna"] == True:
        steps = [
            "s01indexing",
            "s02mapping",
            "s03indexing",
            "s04mapping",
            "s05calling",
            "s06jointcalling",
            "s07merging",
            "s08annotating",
            "s09expressing",
            "s10velocitying",
        ]
    else:
        steps = [
            "s02mapping",
            "s04mapping",
            "s05calling",
            "s06jointcalling",
            "s07merging",
            "s08annotating",
        ]

    conds = {}
    for cond in steps:
        x = 0
        for file in glob.glob(f"{config['tmpdir']}/log/{cond}/*.o"):
            if not _is_ok(file):
                x += 1
        conds[cond] = x
    print(conds)



# -*- coding: utf-8 -*-
#!/usr/bin/env python
# -*- coding: utf-8 -*-


# =========================================================================================
# Written by : Can Kizilkale (CanKizilkale@lbl.gov) and Aydin Buluc (abuluc@lbl.gov)
# Last Update: Jan 31, 2021
# =========================================================================================

"""
Created on Thu Feb 13 19:18:26 2020

@author: Can KIZILKALE
"""

# Approximate Error Correction for phylogenetic tree matrices.
# Can Kizilkale

import argparse
import copy
import itertools
import multiprocessing as mp
import os
import time
from argparse import ArgumentParser
from datetime import datetime
from multiprocessing import Process, Queue

import networkx as nx
import numpy as np
import pandas as pd

os.environ["PATH"] += os.pathsep + "C:/Program Files (x86)/Graphviz2.38/bin/"
np.set_printoptions(threshold=np.inf)

# ---------------------READ ME ---------------------------
# Call this function like:
#
# \python huntress.py "Noisy_matrix_filename" "Output_filename" --nofcpus 8 --algorithmchoice "FPNA" --fn_fpratio 100 --fp_coeff 0.0001 --fn_coeff 0.1
# Output is written on Output_filename.CFMATRIX
#
# The path and name of the the noisy matrix is given in Noisy_matrix_filename
# The reconstructed error free matrix is written in Output_filename with the extension ".CFMatrix"
# The optional inputs are as follows:
# --nofcpus defines the number of cpus to be used for tuning in parallel. Default is 7
# --algorithmchoice defines the version of the algorithm to be used.
#           = "FN" for matrices that only have false negatives
#           = "FPNA" for matrices that have false positives , false negatives and NA (entries that could not be read) entries. These entries must be given as 3 in the input matrix
# --fn_fpratio is the ratio of the weights of 0->1 switches over 1->0 switches that is used by the algorithm to tune the parameters.
# Default=100
# --fp_coeff false positive probability coefficient used for postprocessing.
# Default: 0.0001
# --fn_coeff false negative probability coefficient used for postprocessing.
# Default: 0.1
#
# ------------------- END README ---------------------------------------

prfp = 5


def Reconstruct(
    input_file,
    output_file,
    Algchoice="FPNA",
    auto_tune=1,
    overlapp_coeff=0.3,
    hist_coeff=20,
    postprocessing=0,
    fnfp=100,
    fnc=1,
    postprcoef=5,
    post_fp=0.0001,
    post_fn=0.1,
    n_proc=7,
):
    global prfp
    prfp = postprcoef
    q = Queue()
    Flog = open(output_file + ".LOG", "a+")
    matrix_input = ReadFileNA(input_file)
    #    matrix_input=preproc_row(matrix_input,0.5)
    matrix_input_raw = ReadFasis(input_file)
    matrix_NA_est = Estimated_Matrix(input_file)
    print(np.sum(matrix_input), np.sum(matrix_NA_est), file=Flog)
    h_range = [x for x in range(1, 100, 2)]
    oc_range = [x / 10 for x in range(1, 5)]
    tune_var = [x for x in itertools.product(h_range, oc_range)]

    running_time = 0

    if Algchoice == "FN":
        s_time = time.time()
        matrix_recons = greedyPtreeNew(matrix_input.astype(bool))[1]
        e_time = time.time()
        running_time = e_time - s_time
        WriteTfile(output_file, matrix_recons, input_file)

    if Algchoice == "FPNA" and auto_tune == 0:

        s_time = time.time()
        apprx_ordr = sum(matrix_NA_est)
        matrix_recons = greedyPtreeNA(
            matrix_input.astype(bool), apprx_ordr, overlapp_coeff, hist_coeff
        )[2]
        e_time = time.time()
        running_time = e_time - s_time
        output_file = output_file + "_optH_{}TEMP.CFMatrix".format(hist_coeff)
        WriteTfile(output_file, matrix_recons, input_file)
    #        print(" difference ",np.sum(matrix_recons!=matrix_input))

    if Algchoice == "FPNA" and auto_tune == 1:
        proc_size = np.ceil(len(tune_var) / n_proc).astype(int)
        print(proc_size)
        cpu_range = []
        #        q=Queue()

        for i in range(n_proc):
            s_i = i * proc_size
            e_i = s_i + proc_size
            if e_i > len(tune_var):
                e_i = len(tune_var)
            cpu_range.append(tune_var[s_i:e_i])
        s_time = time.time()
        p = [
            Process(
                target=Auto_fnfp,
                args=(
                    q,
                    cpu_range[i],
                    matrix_input,
                    matrix_NA_est,
                    matrix_input_raw,
                    fnfp,
                    fnc,
                    i,
                    input_file,
                    output_file,
                ),
            )
            for i in range(len(cpu_range))
        ]
        for i in p:
            i.start()
        for i in p:
            i.join()
        ret = []
        while not q.empty():
            print(" Reading the queue ...")
            ret.append(q.get())

        [m_r, d_min] = ret[0]
        matrix_recons = ReadFfile(m_r)
        for i in range(1, len(ret)):
            [m, d] = ret[i]
            print(d_min)
            if d < d_min:
                matrix_recons = ReadFfile(m)
                print(d_min)
                d_min = d

        print("closed", q.empty())

        e_time = time.time()
        running_time = e_time - s_time
        print(running_time)
        output_file = output_file + "_optH_TEMP.CFMatrix"
        WriteTfile(output_file, matrix_recons, input_file)
        print(
            " 1->0 : ",
            np.sum(matrix_recons < matrix_input),
            " 0->1 : ",
            np.sum(matrix_recons > matrix_input_raw),
            " NA->1 : ",
            np.sum(matrix_recons > matrix_input)
            - np.sum(matrix_recons > matrix_input_raw),
        )

    Flog.close()
    postprocess_col(
        input_file, output_file, pfn=post_fn, pfp=post_fp
    )  # Column based post processing, seem to give the best result.

    return running_time


def Auto_fnfp(
    q, tune_ran, m_input, m_NA_est, m_input_raw, fnfp, fnc, procid, in_file, out_file
):  # multiprocessing worker unit, takes a tuning range and tries everything in between
    apprx_ordr = sum(m_NA_est)

    print("An instance of auto tune has started...", procid)
    matrix_recon = greedyPtreeNA(
        m_input.astype(bool), apprx_ordr, tune_ran[0][1], tune_ran[0][0]
    )[2]
    matrix_rec_Temp = deleteNas(m_input_raw, matrix_recon)
    n_10 = np.sum(matrix_rec_Temp < m_input)
    n_01 = np.sum(matrix_rec_Temp > m_input)

    #        distance=np.square(n_10)+n_01
    distance = fnfp * n_10 + fnc * n_01

    h_current = tune_ran[0][0]
    oc_current = tune_ran[0][1]
    for [h_i, overlapp_coeff] in tune_ran:
        matrix_rec_i = greedyPtreeNA(
            m_input.astype(bool), apprx_ordr, overlapp_coeff, h_i
        )[2]
        matrix_rec_Temp = deleteNas(m_input_raw, matrix_rec_i)
        n_10 = np.sum(matrix_rec_Temp < m_input, dtype="int64")
        n_01 = np.sum(matrix_rec_Temp > m_input, dtype="int64")
        #        print("Opt h = ",h_current,"Opt_O = ",oc_current,"H,OC coefficient = ",h_i,overlapp_coeff," 01 switches : ",n_01,"10 switches = ", n_10 , "best :",distance)
        print(procid, h_i, overlapp_coeff)
        #           distance_i=np.square(n_10)+n_01
        distance_i = fnfp * n_10 + fnc * n_01
        #        distance_i=-fnfp**n_10*fnc**n_01
        #            if n_10<np.sum(matrix_recons<matrix_input):
        if distance_i < distance:
            matrix_recon = matrix_rec_i.copy()
            distance = distance_i
            #                WriteTfile(output_file,matrix_recons,input_file)
            h_current = h_i
            oc_current = overlapp_coeff
    print(procid, "th process finished", q.full(), h_current, oc_current)
    output_file = out_file + "_TEMP_{}.CFMatrix".format(procid)
    WriteTfile(output_file, matrix_recon, in_file)
    q.put([output_file, distance])


def deleteNas(M_in, M_out):
    M_o = M_out.copy()
    NA_position = np.argwhere(M_in == 3)
    # print("number of NA : ",len(NA_position))
    for j in NA_position:
        # print(M[j[0],j[1]])
        M_o[j[0], j[1]] = 0

    return M_o


def deletemutations(
    M_in, M_out
):  # Finds the rows which are too far from the input, just replicates them to their closest neighbour.
    x = M_in.shape
    M_return = M_out.copy()
    treshold_cut = 0.5
    dt = np.divide(sum(M_in), sum(M_out)) <= treshold_cut
    for i in range(x[1]):
        if dt[i] == 1:
            M_return[:, i] = np.zeros(x[0])

    return M_return


def precombination(M_in):
    x = M_in.shape
    trsh = 0.22
    M_return = M_in.copy()
    for i in range(x[0]):
        for j in range(i + 1, x[0]):
            rij = np.sum(M_in[i, :] != M_in[j, :]) / np.sum(M_in[i, :] + M_in[j, :])
            print(rij)
            if rij < trsh:
                cupi = M_in[i, :] + M_in[j, :]
                M_return[i, :] = cupi
                M_return[j, :] = cupi
                print("combined ", i, j)
    return M_return


def findclosestinter(i, M_input):
    vec_int = np.zeros(M_input.shape[1]).astype(bool)
    for j in range(M_input.shape[0]):
        if (
            i != j
            and np.sum(M_input[j, :] > M[i, :]) > 0
            and np.sum(M_input[j, :] * M[i, :]) > 0
        ):
            if np.sum(vec_int):
                vec_int = M_input[j, :]
            else:
                vec_int = vec_int * M_input[j, :]
    return vec_int


# Both algorithms take inut matrices of BOOL type.
def greedyPtreeNew(
    M_input,
):  # very greedy algorithm that constructs a ptree matrix from M_inputs by adding 1's
    # M_input has to be of type BOOLEAN !
    # Returns
    # M_copy(reconstructed matrix),
    # bret (list of positions of assumed false negatives)
    #    M_apprx=greedy_row_rec(M_input)[1]
    M_copy = M_input.copy()

    ISet1 = np.argsort(sum(M_copy))
    #    ISet1=np.argsort(sum(M_apprx))
    #    ISet1=np.argsort(sum(M_rowdene))
    ISet = []
    for i in range(M_copy.shape[1]):
        ISet.append(ISet1[i])

    bret = []  # Location of detected false negatives
    print(M_copy.shape, len(ISet))
    while len(ISet) > 1:
        # print("pivoting column", i,bret)
        pivot_index = ISet[-1]  # index of the pivot vector
        Sremaining = ISet.copy()
        pivot_vector = M_copy[
            :, pivot_index
        ]  # vector used for pivoting the current iteration
        cum_vector = np.copy(pivot_vector)  # holds the union vector
        while_cont = 1

        while while_cont == 1:  # main loop
            while_cont = 0

            for j in Sremaining:
                cap_j = (
                    cum_vector * M_copy[:, j]
                )  # intersection of the pivot and the jth column

                if np.any(
                    cap_j
                ):  # continue as long as there is a column having non-empty intersection
                    cum_vector = cum_vector + M_copy[:, j]
                    while_cont = 1
                    Sremaining.remove(j)
        #                    print("pivot ", pivot_index, j, sum(cap_j),sum(cum_vector),sum(M_copy[:,j]))
        #        print(len(Si))
        #        print(sum(pivot_vector),sum(cum_vector))
        M_copy[:, pivot_index] = cum_vector
        ISet.remove(pivot_index)
        bret = np.argwhere(M_copy.astype(int) > M_input.astype(int))
    return [bret, M_copy]


def greedyPtreeNA(
    M_input, approx_order, oc, hc
):  # Modified Greedy algorithm for NA and false positive values.
    # Matrix M_input has to be BOOLEAN for the code to work right
    # Returns:
    # M_copy(reconstructed matrix),
    # bret (list of positions of assumed false negatives)
    # pret (list of approximate false positives)
    overlap_coeff = oc  # 0.1
    hist_coeff = hc  # 25

    M_copy = M_input.copy()
    #    ISet1=np.argsort(sum((M_copy.T.dot(M_copy)>0).T))    # Set of indices of columns sorted wrt. number of overlapping vectors
    #    ISet1=np.argsort(sum(M_copy))         # Set of indices of columns sorted wrt. the number of ones they contain
    # ISet1=np.argsort(sum(M_NA))
    ISet1 = np.argsort(approx_order)
    ISet = []
    for i in range(M_copy.shape[1]):
        ISet.append(ISet1[i])

    bret = []  # Location of detected false negatives
    pret = []  # Location of detected false positives

    while len(ISet) > 1:
        #        pivot_index=ISet[np.argmax(sum((M_copy[:,ISet].T.dot(M_copy[:,ISet])>0).T))]
        pivot_index = ISet[-1]  # index of the pivot vector
        Sremaining = ISet.copy()  # set of indices that are not included in the union
        pivot_vector = M_copy[
            :, pivot_index
        ]  # vector used for pivoting the current iteration
        cum_vector = np.copy(pivot_vector)  # holds the union vector
        while_cont = 1
        cum_hist = np.zeros(M_input.shape[0])  # holds the histogram for the union

        while while_cont == 1:  # Continue uniting vectors until no candidate remains
            while_cont = 0
            for j in Sremaining:
                cap_i = pivot_vector * M_copy[:, j]
                min_vec_size = np.min([np.sum(M_copy[:, j]), np.sum(pivot_vector)])
                cap_size = np.sum(cap_i)
                #               print(np.sum(M_copy[:,j]),np.sum(pivot_vector),np.sum(cap_i),cap_size, min_vec_size,np.floor(cap_size/min_vec_size))

                if (
                    cap_size / min_vec_size > overlap_coeff
                ):  # we check if the columns have a meaningful overlap
                    cum_hist = cum_hist + M_copy[:, j].astype(int)
                    while_cont = 1  # found an overlapping vector so we keep on going
                    Sremaining.remove(j)  # united vector is removed from the search set

        cnumT = np.floor(
            cum_hist.max() / hist_coeff
        )  # the elements that repeated few times are considered to be false positives
        cum_vector = cum_hist > cnumT

        pivot_est = pivot_index

        ncap = np.sum(cum_vector * M_copy[:, pivot_est])
        #        ndel=np.sum(M_copy[:,j]!=capj)
        for j in ISet:  # clean up the false positives wrt. the established pivot
            capj = cum_vector * M_copy[:, j]  # intersection of union with column j
            #            difj=M_copy[:,j]!=capj                        # difference of column j from the union
            difj = M_copy[:, j] > capj  # difference of column j from the union
            if np.sum(capj) > np.sum(difj):
                M_copy[:, j] = capj

            else:
                M_copy[:, j] = difj

        M_copy[
            :, pivot_index
        ] = cum_vector  # correcting the false negatives in the pivot
        ISet.remove(pivot_index)  # removing the pivot from the search space
    #        print(len(ISet),np.sum(M_copy[:,pivot_index]),np.sum(M_input[:,pivot_index]))

    bret = np.argwhere(M_copy.astype(int) > M_input.astype(int))
    pret = np.argwhere(np.argwhere(M_copy.astype(int) < M_input.astype(int)))
    return [bret, pret, M_copy]


def ReadFfile(filename):  # reads a matrix from file and returns it in BOOL type
    df = pd.read_csv(filename, sep="\t", index_col=0)
    M = df.values.astype(bool)
    return M


def ReadFile(filename):  # reads a matrix from file and returns it in BOOL type
    df = pd.read_csv(filename, sep="\t", index_col=0)
    M = df.values
    return M


def ReadFileNA(filename):  # reads the file and fills the NA with 0's.
    df = pd.read_csv(filename, sep="\t", index_col=0)
    M = df.values

    NA_position = np.argwhere(M == 3)
    # print("number of NA : ",len(NA_position))
    for j in NA_position:
        # print(M[j[0],j[1]])
        M[j[0], j[1]] = 0

    # print(sum(sum(M)))
    return M.astype(bool)


def Estimated_Matrix(
    filename,
):  # Creates an estimate of the matrix such that each element is given the expectation wrt the column 1/0 frequencies.
    df = pd.read_csv(filename, sep="\t", index_col=0)
    M = df.values.astype(float)

    for i in range(M.shape[1]):
        if np.sum(M[:, i] != 3) == 0:
            one_ratio = 0
        else:
            one_ratio = np.sum(M[:, i] == 1) / np.sum(M[:, i] != 3)
        for j in range(M.shape[0]):
            if M[j, i] == 3:
                M[j, i] = one_ratio

    return M


def WriteTfile(
    filename, matrix, filename2
):  # writes matrix output as an integer matrix
    df_input = pd.read_csv(filename2, sep="\t", index_col=0)
    matrix_output = matrix.astype(int)
    df_output = pd.DataFrame(matrix_output)
    df_output.columns = df_input.columns
    df_output.index = df_input.index
    df_output.index.name = "cellIDxmutID"
    df_output.to_csv(filename, sep="\t")


##########################################################################3
############# TESTING FUNCTIONS ##########################################
def isPtree(matrix_in):  # brute force check if matrix_in is a pTree
    M = matrix_in.astype(bool)
    for j in range(M.shape[1]):
        for i in range(j, M.shape[1]):
            cap = M[:, i] * M[:, j]
            cap_size = np.sum(cap)
            Mi_size = np.sum(M[:, i])
            Mj_size = np.sum(M[:, j])
            if cap_size != 0:
                if cap_size != Mi_size:
                    if cap_size != Mj_size:
                        return False

    print("Seems to be a PTree ...")
    return True


def compareAD(M1, M2):  # M1 is the ground truth
    error_pairs = []
    n_adpairs = 0
    for i in range(M1.shape[1]):
        #        print(i)
        for j in range(i, M1.shape[1]):
            cap1 = M1[:, i] * M1[:, j]
            cap2 = M2[:, i] * M2[:, j]
            if np.sum(cap1) > 0 and np.sum(M1[:, i]) != np.sum(M1[:, j]):
                n_adpairs = n_adpairs + 1
                if np.sum(cap2) == 0:
                    error_pairs.append([i, j])
                else:
                    if np.sum(M1[:, j]) > np.sum(M1[:, i]) and np.sum(
                        M2[:, j]
                    ) <= np.sum(M2[:, i]):
                        error_pairs.append([i, j])
                    else:
                        if np.sum(M1[:, i]) > np.sum(M1[:, j]) and np.sum(
                            M2[:, i]
                        ) <= np.sum(M2[:, j]):
                            error_pairs.append([i, j])
                        # print(i,j,sum(M1[:,i]),sum(M1[:,j]),sum(M2[:,i]),sum(M2[:,j]))
    print(
        "Number of AD pairs = ",
        n_adpairs,
        "errors : ",
        len(error_pairs),
        "AD score = ",
        1 - len(error_pairs) / n_adpairs,
    )
    return error_pairs


def compareDF(M_orj, M_rec):
    error_pairs = []
    d_pairs = 0
    for i in range(M_orj.shape[1]):
        for j in range(i, M_orj.shape[1]):
            cap1 = M_orj[:, i] * M_orj[:, j]
            cap2 = M_rec[:, i] * M_rec[:, j]
            if np.sum(cap1) == 0:
                d_pairs = d_pairs + 1
                if np.sum(cap2) > 0:
                    error_pairs.append([i, j])
    print(
        "Number of Diff pairs = ",
        d_pairs,
        "errors :",
        len(error_pairs),
        "score :",
        1 - len(error_pairs) / d_pairs,
    )
    return


def ReadFasis(filename):  # reads a matrix from file and returns it in BOOL type
    df = pd.read_csv(filename, sep="\t", index_col=0)
    M = df.values
    return M


def compute_fnfp(M_n, M_r):
    n_01 = 0
    n_10 = 0
    n_1 = 0
    n_0 = 0
    for x in np.argwhere(M_n < 3):
        if M_r[x[0], x[1]] == 0:
            n_0 = n_0 + 1
        if M_r[x[0], x[1]] == 1:
            n_1 = n_1 + 1
        if M_n[x[0], x[1]] > M_r[x[0], x[1]]:

            n_10 = n_10 + 1

        if M_n[x[0], x[1]] < M_r[x[0], x[1]]:
            n_01 = n_01 + 1
            n_1 = n_1 + 1
    print("computed fn :", n_01 / n_1, " fp : ", n_10 / n_0)
    return [n_01 / n_1, n_10 / n_0]


def find_dist(node_piv, M_samples):
    #    distances=np.zeros(M_nodes.shape[0])
    #    for i in range(M_nodes.shape[0]):
    ##        distances[i]=np.sum(M_nodes[i,:]!=node_piv)
    ##        distances[i]=np.sum(M_nodes[i,:]<node_piv)
    ##        distances[i]=np.linalg.norm(M_nodes[i,:]-node_piv)
    #        distances[i]=0
    #        d_10=0
    #        d_01=0
    #        for j in range(M_nodes.shape[1]):
    #
    #            if node_piv[j]!=3:
    #                if node_piv[j]>M_nodes[i,j]:
    #                    d_10=d_10+1
    #                if node_piv[j]<M_nodes[i,j]:
    #                    d_01=d_01+1
    #        distances[i]= np.square(d_10) +d_01
    ##            else:
    ##                distances[i]=1-M_nodes[i,j]+distances[i]
    M_nodes = M_samples.copy()
    for j in range(M_nodes.shape[1]):
        if node_piv[j] == 3:
            node_piv[j] = 0
            M_nodes[:, j] = 0 * M_nodes[:, j]
    distances = np.zeros(M_nodes.shape[0])
    for i in range(M_nodes.shape[0]):
        d_10 = np.sum(node_piv > M_nodes[i, :], dtype="int64")
        d_01 = np.sum(node_piv < M_nodes[i, :], dtype="int64")
        distances[i] = np.square(d_10) + d_01

    return distances


def find_dist_col(node_piv, M_samples):
    M_nodes = M_samples.copy()
    for j in range(M_nodes.shape[0]):
        if node_piv[j] == 3:
            node_piv[j] = 0
            M_nodes[j, :] = 0 * M_nodes[j, :]
    distances = np.zeros(M_nodes.shape[1])
    for i in range(M_nodes.shape[1]):
        d_10 = np.sum(node_piv > M_nodes[:, i], dtype="int64")
        d_01 = np.sum(node_piv < M_nodes[:, i], dtype="int64")
        #        distances[i]=np.square(d_10) + d_01
        #        distances[i]=d_10 + d_01
        distances[i] = prfp * d_10 + d_01
        #        distances[i]=np.sum(M_nodes[:,i]*node_piv)/np.sqrt(np.sum(M_nodes[:,i])*np.sum(node_piv))
        distances[i] = -((0.005) ** d_10) * (0.1) ** d_01
    return distances


def closest_matrix(M_input, M_nodes, M_rec):
    M_out = M_input.copy()
    for i in range(M_input.shape[0]):
        pivot_v = M_input[i, :]
        distance_i = find_dist(pivot_v, M_nodes)

        min_index = np.argmin(distance_i)
        #        print("Current distance ",np.sum(M_input[i,:]>M_rec[i,:]),"minimum possible ",np.min(distance_i),min_index)
        M_out[i, :] = M_nodes[min_index, :]
    #        if np.sum(M_input[i,:]>M_rec[i,:])>2*np.min(distance_i):
    #            min_index=np.argmin(distance_i)
    #            M_out[i,:]=M_nodes[min_index,:]
    #        else:
    #            M_out[i,:]=M_rec[i,:]
    #        print("M input :", sum(M_input[i,:]),sum(M_out[i,:]))
    return M_out


def closest_matrix_col(M_input, M_nodes, M_rec):
    M_out = M_input.copy()
    for i in range(M_input.shape[1]):
        pivot_v = M_input[:, i]
        distance_i = find_dist_col(pivot_v, M_nodes)

        min_index = np.argmin(distance_i)
        #        print("Old New difference ",np.sum(M_nodes[:,i]!=M_nodes[:,min_index]),i)
        M_out[:, i] = M_nodes[:, min_index]

    return M_out


def postprocess_col(input_file, out_file, pfn, pfp):
    s = time.time()

    M_noisy = ReadFasis(input_file)
    M_n_copy = M_noisy.copy()
    M_nds = ReadFfile(out_file)

    Mtemp = c_m_col(ReadFasis(input_file), M_nds, pc_fn=pfn, pc_fp=pfp)
    Mtemp2 = Mtemp.copy()
    #    Mtemp=c_m_row(ReadFasis(input_file),Mtemp,1)
    d10min = np.sum(Mtemp < (M_noisy == 1))
    d10c = d10min
    imp = 1
    while imp:
        Mtemp2 = c_m_row(ReadFasis(input_file), Mtemp2, pc_fn=pfn, pc_fp=pfp)
        Mtemp2 = c_m_col(ReadFasis(input_file), Mtemp2, pc_fn=pfn, pc_fp=pfp)

        #        Mtemp=c_m_col(ReadFasis(input_file),Mtemp,1)
        d10c = np.sum(Mtemp2 < (M_noisy == 1))
        print(d10c)
        if d10c < d10min:
            d10min = d10c
            Mtemp = Mtemp2.copy()
        else:
            imp = 0

    #    M_postprocessed=closest_matrix_col(M_noisy,M_nds,ReadFfile(out_file))
    M_postprocessed = Mtemp
    processed_file = out_file[:-13] + ".CFMatrix"
    #    print("Writing to file ",processed_file,file=Flog)
    WriteTfile(processed_file, M_postprocessed, input_file)
    #    print(np.sum(M_noisy),np.sum(M_noisy==1))
    e = time.time()
    print(
        "Postprocessed 1->0 : ",
        np.sum(M_postprocessed < (M_n_copy == 1)),
        " 0->1 : ",
        np.sum(M_postprocessed > M_n_copy),
        " NA->1 : ",
        np.sum(M_postprocessed > (M_n_copy == 1)) - np.sum(M_postprocessed > M_n_copy),
    )
    print("Post processing time : ", e - s)


#    draw_tree(processed_file)


def preproc_row(M_o, c=0.8):
    M_pre = M_o.copy()
    for i in range(M_o.shape[0]):
        for j in range(i + 1, M_o.shape[0]):
            prdct = np.sum(M_o[i, :] * M_o[j, :]) / np.sqrt(
                np.sum(M_o[i, :]) * np.sum(M_o[j, :])
            )
            if prdct > c:
                print(i, j)
                M_pre[i, :] = M_o[i, :] + M_o[j, :]
                M_pre[j, :] = M_o[i, :] + M_o[j, :]
    return M_pre


def f_d_col(node_piv, M_samples, p_fp=0.005, p_fn=0.1):
    M_nodes = M_samples.copy()
    D10 = (M_nodes.T == 0).astype(int).dot(node_piv == 1)
    D11 = (M_nodes.T == 1).astype(int).dot(node_piv == 1)
    D00 = (M_nodes.T == 0).astype(int).dot(node_piv == 0)
    D01 = (M_nodes.T == 1).astype(int).dot(node_piv == 0)
    distances = np.zeros(M_nodes.shape[1])
    distances = -np.multiply(np.power(p_fp, D10), np.power(1 - p_fp, D00))
    distances = np.multiply(distances, np.power(p_fn, D01))
    distances = np.multiply(distances, np.power(1 - p_fn, D11))
    #    for i in range(M_nodes.shape[1]):
    #        d_10=D10[i]
    #        d_11=D11[i]
    #        d_00=D00[i]
    #        d_01=D01[i]
    ##        d_10=np.sum((M_nodes[:,i]==0)*(node_piv==1),dtype='int64')
    ##        d_11=np.sum((M_nodes[:,i]==1)*(node_piv==1),dtype='int64')
    ##        d_00=np.sum((M_nodes[:,i]==0)*(node_piv==0),dtype='int64')
    ##        d_01=np.sum((M_nodes[:,i]==1)*(node_piv==0),dtype='int64')
    #
    #        distances[i]=-(p_fp)**d_10*(1-p_fp)**d_00*(p_fn)**d_01*(1-p_fn)**d_11
    return distances


def c_m_col(M_input, M_nodes, pc_fp=0.0001, pc_fn=0.1):
    M_out = M_input.copy()

    for i in range(M_input.shape[1]):
        pivot_v = M_input[:, i]

        distance_i = f_d_col(pivot_v, M_nodes, p_fn=pc_fn, p_fp=pc_fp)

        min_index = np.argmin(distance_i)
        #        print("Old New difference ",np.sum(M_nodes[:,i]!=M_nodes[:,min_index]),i)
        M_out[:, i] = M_nodes[:, min_index]

    return M_out


def f_d_row(node_piv, M_samples, p_fp=0.005, p_fn=0.1):
    M_nodes = M_samples.copy()

    distances = np.zeros(M_nodes.shape[0])
    D10 = (M_nodes == 0).astype(int).dot(node_piv == 1)
    D11 = (M_nodes == 1).astype(int).dot(node_piv == 1)
    D00 = (M_nodes == 0).astype(int).dot(node_piv == 0)
    D01 = (M_nodes == 1).astype(int).dot(node_piv == 0)

    distances = -np.multiply(np.power(p_fp, D10), np.power(1 - p_fp, D00))
    distances = np.multiply(distances, np.power(p_fn, D01))
    distances = np.multiply(distances, np.power(1 - p_fn, D11))
    #    for i in range(M_nodes.shape[0]):
    #        d_10=D10[i]
    #        d_11=D11[i]
    #        d_00=D00[i]
    #        d_01=D01[i]
    ##
    ##        d_10=np.sum((M_nodes[i,:]==0)*(node_piv==1),dtype='int64')
    ##        d_11=np.sum((M_nodes[i,:]==1)*(node_piv==1),dtype='int64')
    ##        d_00=np.sum((M_nodes[i,:]==0)*(node_piv==0),dtype='int64')
    ##        d_01=np.sum((M_nodes[i,:]==1)*(node_piv==0),dtype='int64')
    #
    #        distances[i]=-(p_fp)**d_10*(1-p_fp)**d_00*(p_fn)**d_01*(1-p_fn)**d_11
    return distances


def c_m_row(M_input, M_nodes, pc_fp=0.0001, pc_fn=0.1):
    M_out = M_input.copy()

    for i in range(M_input.shape[0]):
        pivot_v = M_input[i, :]
        distance_i = f_d_row(pivot_v, M_nodes, p_fn=pc_fn, p_fp=pc_fp)

        min_index = np.argmin(distance_i)
        #        print("Old New difference ",np.sum(M_nodes[:,i]!=M_nodes[:,min_index]),i)
        M_out[i, :] = M_nodes[min_index, :]

    return M_out


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("inputfile")
    parser.add_argument("outputfile")
    parser.add_argument("--nofcpus", default=7, type=int, nargs="?")
    parser.add_argument("--algorithmchoice", default="FPNA", nargs="?")
    parser.add_argument("--fn_fpratio", default=51, type=int, nargs="?")
    parser.add_argument("--fp_coeff", default=0.00001, type=float, nargs="?")
    parser.add_argument("--fn_coeff", default=0.1, type=float, nargs="?")
    args = parser.parse_args()

    fn_conorm = 0.1
    fp_conorm = fn_conorm * args.fp_coeff / args.fn_coeff
    #    fnfp_conorm = fn_conorm / fp_conorm
    #    Reconstruct(args.inputfile,args.outputfile,Algchoice=args.algorithmchoice,n_proc=args.nofcpus,fnfp=args.fn_fpratio,post_fn=args.fn_coeff,post_fp=args.fp_coeff)
    Reconstruct(
        args.inputfile,
        args.outputfile,
        Algchoice=args.algorithmchoice,
        n_proc=args.nofcpus,
        fnfp=args.fn_fpratio,
        post_fn=fn_conorm,
        post_fp=fp_conorm,
    )



import anndata as ad
import numpy as np
import pandas as pd
from scipy import stats

import trisicell as tsc


def count_mut_in_a_gene(adata, cut_off=0):
    x = pd.Series([".".join(x.split(".")[:-4]) for x in adata.var_names]).value_counts()
    return x[x > cut_off]


def remove_indels(adata):
    snps = []
    for x in adata.var_names:
        ens, gene, chrom, pos, ref, alt = tsc.ul.split_mut(x)
        if ("-" not in ref) and ("-" not in alt):
            snps.append(x)
    keep_mut_by_list(adata, snps)


def most_frequent_muts(adata):
    G = adata.to_df(layer="genotype")
    V = adata.to_df(layer="mutant")
    T = adata.to_df(layer="total")
    s1 = ((G == 1) | (G == 3)).sum(axis=0).rename("n_cells")
    s2 = V.sum(axis=0).rename("overall_sum_mutant_coverage")
    s3 = T.sum(axis=0).rename("overall_sum_total_coverage")
    return pd.concat([s1, s2, s3], axis=1).sort_values(
        by=["n_cells", "overall_sum_mutant_coverage", "overall_sum_total_coverage"],
        ascending=False,
    )


def muts_in_a_subset_of_cells(adata, alist):
    G = adata[alist].layers["genotype"]
    return adata.var_names[((G == 1) | (G == 3)).sum(axis=0) > 0]


def n_muts_per_cell_with_vaf(adata):
    G = adata.layers["genotype"]
    n_muts = ((G == 1) | (G == 3)).sum(axis=1)
    V = adata.layers["mutant"]
    T = adata.layers["total"]

    vaf = np.zeros(G.shape)
    vaf = np.divide(V, T, where=T != 0)
    vaf[(G == 1) | (G == 3)] = 0
    vaf = np.nanmean(np.where(vaf != 0, vaf, np.nan), axis=1)

    n_mutant = V
    n_mutant[(G == 1) | (G == 3)] = 0
    n_mutant = np.nanmean(np.where(n_mutant != 0, n_mutant, np.nan), axis=1)

    n_total = T
    n_total[(G == 1) | (G == 3)] = 0
    n_total = np.nanmean(np.where(n_total != 0, n_total, np.nan), axis=1)
    return pd.DataFrame(
        {
            "n_muts": n_muts,
            "avg_vaf": vaf,
            "avg_n_mutant": n_mutant,
            "avg_n_total": n_total,
        },
        index=adata.obs_names,
    )


def get_vaf(adata):
    V = adata.to_df(layer="mutant")
    T = adata.to_df(layer="total")
    return V / T


def get_vaf_with_list(adata, alist):
    x = np.intersect1d(alist, adata.var_names)
    y = np.setdiff1d(alist, adata.var_names)
    df = get_vaf(adata[:, x])
    df[y] = None
    return df.transpose()


def bulk_total_mutant_vaf(adata):
    V = adata.to_df(layer="mutant")
    T = adata.to_df(layer="total")
    s1 = T.sum(axis=0).rename("total")
    s2 = V.sum(axis=0).rename("mutant")
    s3 = (s2 / s1).rename("vaf")
    return pd.concat([s1, s2, s3], axis=1)


def filter_mut_bulk_mutant_and_vaf_greater_than(adata, min_vaf=0.2, min_mutant=20):
    df = bulk_total_mutant_vaf(adata)
    good_muts = df[(df["vaf"] >= min_vaf) & (df["mutant"] >= min_mutant)].index
    keep_mut_by_list(adata, good_muts)


def filter_mut_total_coverage_must_be_greater_than(adata, min_total=7, min_cells=1):
    T = adata.layers["total"]
    good_muts = (T >= min_total).sum(axis=0) >= min_cells
    keep_mut_by_list(adata, adata.var_names.to_numpy()[good_muts])


def filter_mut_vaf_greater_than_coverage_mutant_greater_than(
    adata, min_vaf=0.2, min_coverage_mutant=20, min_cells=1
):
    V = adata.to_df(layer="mutant")
    T = adata.to_df(layer="total")

    # good_muts_1 = (V >= min_coverage_mutant).sum(axis=0) >= min_cells
    df = V / T
    good_muts = ((df >= min_vaf) & (V >= min_coverage_mutant)).sum(axis=0) >= min_cells
    # a = adata.var_names.to_numpy()[good_muts_1]
    # b = adata.var_names.to_numpy()[good_muts_2]
    # good_muts = np.intersect1d(a, b)
    keep_mut_by_list(adata, adata.var_names.to_numpy()[good_muts])


def filter_mut_coverage_mutant_must_be_greater_than(
    adata, min_coverage_mutant=3, min_cells=1
):
    V = adata.layers["mutant"]
    good_muts = (V >= min_coverage_mutant).sum(axis=0) >= min_cells
    keep_mut_by_list(adata, adata.var_names.to_numpy()[good_muts])


def _helper(sub_adata):
    genotype = 2 * np.ones(sub_adata.shape[1])
    mutant = sub_adata.layers["mutant"].sum(axis=0)
    total = sub_adata.layers["total"].sum(axis=0)

    genotype[total == mutant] = 3
    genotype[total > mutant] = 1
    genotype[mutant == 0] = 0
    genotype[total == 0] = 2

    return genotype, mutant, total


def merge_cells_using(adata, using):
    genotypes = []
    mutants = []
    totals = []
    indices = []
    for index, subgroup in adata.obs.groupby(using):
        genotype, mutant, total = _helper(adata[subgroup.index, :])
        indices.append(index)
        genotypes.append(genotype)
        mutants.append(mutant)
        totals.append(total)

    adata_merged = ad.AnnData(
        X=3 * np.ones((len(indices), adata.shape[1])),
        obs=pd.DataFrame.from_dict({"cells": indices}).set_index("cells"),
        var=adata.var,
        layers={
            "total": np.array(totals),
            "mutant": np.array(mutants),
            "genotype": np.array(genotypes),
        },
    )
    return adata_merged


def merge_clusters_and_call_mutations(adata, n_clusters, min_n_cells, using):
    tsc.pp.build_scmatrix(adata)
    dist = tsc.ul.calc_dist_by_l1_ignore_na(adata.X)
    clusters = tsc.ul.calc_hclust(dist, adata.obs_names, method="ward")

    cluster = clusters[n_clusters]
    count = cluster.value_counts()
    cluster = cluster[cluster.isin(count[count >= min_n_cells].index)]
    cluster = cluster.apply(lambda x: f"G{x}")

    adata = adata[cluster.index, :]
    adata.obs[using] = cluster

    genotypes = []
    mutants = []
    totals = []
    indices = []
    for index, subgroup in adata.obs.groupby(adata.obs[using]):
        genotype, mutant, total = _helper(adata[subgroup.index, :])
        indices.append(index)
        genotypes.append(genotype)
        mutants.append(mutant)
        totals.append(total)

    adata_merged = ad.AnnData(
        X=3 * np.ones((len(indices), adata.shape[1])),
        obs=pd.DataFrame.from_dict({"cells": indices}).set_index("cells"),
        var=adata.var,
        layers={
            "total": np.array(totals),
            "mutant": np.array(mutants),
            "genotype": np.array(genotypes),
        },
    )
    return adata_merged, adata


def filter_mut_mutant_must_present_in_at_least(adata, min_cells=2):
    """Remove mutations based on the mutant status.

    This function removes mutations that don't have the status of mutant
    in at least `min_cells` cells.

    Parameters
    ----------
    adata : :class:`anndata.AnnData`
        The input readcount data.
    min_cells : :obj:`int`, optional
        Minimum number of cells, by default 1
    """

    G = adata.layers["genotype"]
    good_muts = ((G == 1) | (G == 3)).sum(axis=0) >= min_cells
    keep_mut_by_list(adata, adata.var_names.to_numpy()[good_muts])


def filter_mut_mutant_in_groups_must_present_in_at_least(
    adata, min_groups=2, group_name="group", min_cells_within_group=1
):
    all_good_muts = []
    for index, subgroup in adata.obs.groupby(group_name):
        G = adata[subgroup.index, :].layers["genotype"]
        good_muts = ((G == 1) | (G == 3)).sum(axis=0) >= min_cells_within_group
        all_good_muts.append(good_muts)
    all_good_muts = np.array(all_good_muts)
    good_muts = all_good_muts.sum(axis=0) >= min_groups
    keep_mut_by_list(adata, adata.var_names.to_numpy()[good_muts])


def filter_mut_reference_must_present_in_at_least(adata, min_cells=1):
    """Remove mutations based on the wild-type status.

    This function removes mutations that don't have the status of wild-type
    in at least `min_cells` cells. Note that mutations that are present in all cells
    will not be filtered out by this function.

    Parameters
    ----------
    adata : :class:`anndata.AnnData`
        The input readcount data.
    min_cells : :obj:`int`, optional
        Minimum number of cells, by default 1
    """

    G = adata.layers["genotype"]
    good_muts = ((G == 0).sum(axis=0) >= min_cells) | (
        ((G == 1) | (G == 3)).sum(axis=0) == adata.shape[0]
    )
    keep_mut_by_list(adata, adata.var_names.to_numpy()[good_muts])


def filter_mut_reference_in_groups_must_present_in_at_least(
    adata, min_groups=2, group_name="group", min_cells_within_group=1
):
    all_good_muts = []
    for index, subgroup in adata.obs.groupby(group_name):
        G = adata[subgroup.index, :].layers["genotype"]
        good_muts = ((G == 0).sum(axis=0) >= min_cells_within_group) & (
            ((G == 1) | (G == 3)).sum(axis=0) == 0
        )
        all_good_muts.append(good_muts)
    all_good_muts = np.array(all_good_muts)
    good_muts = all_good_muts.sum(axis=0) >= min_groups
    keep_mut_by_list(adata, adata.var_names.to_numpy()[good_muts])


def filter_mut_by_confidence_score(adata, min_p_value):
    T = adata.layers["total"]
    G = adata.layers["genotype"]

    # x = 0
    mut_to_remove = []
    for j, mut in enumerate(adata.var_names):
        mask = (G[:, j] == 1) | (G[:, j] == 3)
        g1 = T[:, j][mask]
        g2 = T[:, j][~mask]
        p_value = stats.ranksums(g1, g2)[1]
        if p_value < min_p_value:
            # x += 1
            mut_to_remove.append(mut)
            # tsc.logg.info(p_value, g1, g2)
        # if x == 10:
        #     break
    remove_mut_by_list(adata, mut_to_remove)


def remove_mut_by_list(adata, alist):
    """Remove a list of mutations from the data.

    Parameters
    ----------
    adata : :class:`anndata.AnnData`
        The input readcount data.
    alist : :obj:`list`
        A list of mutations.
    """

    adata._inplace_subset_var(np.setdiff1d(adata.var_names, alist))
    tsc.logg.info(f"Matrix with n_obs × n_vars = {adata.shape[0]} × {adata.shape[1]}")


def keep_mut_by_list(adata, alist):
    adata._inplace_subset_var(np.intersect1d(alist, adata.var_names))
    tsc.logg.info(f"Matrix with n_obs × n_vars = {adata.shape[0]} × {adata.shape[1]}")


def remove_cell_by_list(adata, alist):
    """Remove a list of cells from the data.

    Parameters
    ----------
    adata : :class:`anndata.AnnData`
        The input readcount data.
    alist : :obj:`list`
        A list of cells.
    """

    adata._inplace_subset_obs(np.setdiff1d(adata.obs_names, alist))
    tsc.logg.info(f"Matrix with n_obs × n_vars = {adata.shape[0]} × {adata.shape[1]}")


def keep_cell_by_list(adata, alist):
    adata._inplace_subset_obs(np.intersect1d(alist, adata.obs_names))
    tsc.logg.info(f"Matrix with n_obs × n_vars = {adata.shape[0]} × {adata.shape[1]}")


def get_germline_variants(adata, normal_cells, min_vaf=0.1, min_coverage=10):
    if not isinstance(normal_cells, list):
        raise "normal_cells must be a list"

    V = adata[normal_cells, :].to_df(layer="mutant")
    T = adata[normal_cells, :].to_df(layer="total")
    G = adata[normal_cells, :].to_df(layer="genotype")
    df = V / T
    good_muts = ((df >= min_vaf) & (V >= min_coverage) & ((G == 1) | (G == 3))).sum(
        axis=0
    ) >= 1
    return adata.var_names.to_numpy()[good_muts]


def build_scmatrix(adata):
    G = adata.layers["genotype"]
    adata.X[G == 0] = 0
    adata.X[(G == 1) | (G == 3)] = 1
    adata.X[G == 2] = 3
    adata.X = adata.X.astype("int")

    # M = adata.layers["mutant"]
    # T = adata.layers["total"]
    # adata.X[T != -1] = 3
    # adata.X[T > 0] = 0
    # adata.X[M > 0] = 1
    # adata.X = adata.X.astype("int")


def to_dendro(adata, filepath):
    Z = adata.layers["genotype"]
    Z[(Z == 1) | (Z == 3)] = 1
    Z[Z == 2] = 0
    X = adata.layers["mutant"]
    N = adata.layers["total"]
    pd.DataFrame(Z).to_csv(filepath + "/Z.csv")
    pd.DataFrame(X).to_csv(filepath + "/X.csv")
    pd.DataFrame(N).to_csv(filepath + "/N.csv")


def calc_dist_by_dendro(adata):
    T = adata.layers["total"]
    V = adata.layers["mutant"]
    G = adata.layers["genotype"]
    G[(G == 1) | (G == 3)] = 1
    G[G == 2] = 0
    dist, bad_muts = tsc.ul.calc_dist_by_dendro(T, V, G)
    remove_mut_by_list(adata, bad_muts)
    tsc.logg.info(f"{sum(bad_muts)} mutations filtered")
    return dist


def calc_relative_expression(adata, threshold):
    tpm = adata.to_df(layer="tpm")
    e_ij = np.log2(tpm / 10 + 1)

    # bad_cells = (tpm > 0).sum(axis=1) < 3000
    # tpm = tpm.loc[~bad_cells,:]
    # print(f'{sum(bad_cells)} cells filtered based on n_genes less than 3000')

    # bad_cells = e_ij[genes['housekeeping']].mean(axis=1) < 2.5
    # tpm = tpm.loc[~bad_cells,:]
    # e_ij = e_ij.loc[~bad_cells,:]
    # print(f'{sum(bad_cells)} cells filtered based on housekeeping less than 2.5')

    # bad_cells = e_ij[genes['mitochondria']].mean(axis=1) > 2.5
    # tpm = tpm.loc[~bad_cells,:]
    # e_ij = e_ij.loc[~bad_cells,:]
    # print(f'{sum(bad_cells)} cells filtered based on mitochondria greater than 2.5')

    bad_genes = (np.log2(tpm.mean(axis=0) + 1) < threshold).values
    e_ij = e_ij.loc[:, ~bad_genes]
    adata._inplace_subset_var(adata.var_names.to_numpy()[~bad_genes])
    tsc.logg.info(
        f"{sum(bad_genes)} genes filtered based on average accross cells less than"
        f" {threshold}"
    )

    tsc.logg.info(f"{adata.shape[0]} cells and {adata.shape[1]} genes retained")
    re_ij = e_ij - e_ij.groupby(lambda x: x.split("_")[0]).transform("mean")
    adata.layers["relative"] = re_ij.values


def add_vaf_to_layers(adata):
    V = adata.to_df(layer="mutant")
    T = adata.to_df(layer="total")
    df = V / T
    adata.layers["vaf"] = df


def statistics(adata):
    G = adata.layers["genotype"]
    t = adata.shape[0] * adata.shape[1]
    a = (G == 0).sum().sum()
    b = (G == 1).sum().sum()
    c = (G == 2).sum().sum()
    d = (G == 3).sum().sum()
    tsc.logg.info(f"size = {adata.shape[0]} × {adata.shape[1]}")
    tsc.logg.info(f"    HOM_REF = {a:6d} ({100*a/t:2.1f}%)")
    tsc.logg.info(f"    HET     = {b:6d} ({100*b/t:2.1f}%)")
    tsc.logg.info(f"    HOM_ALT = {d:6d} ({100*d/t:2.1f}%)")
    tsc.logg.info(f"    UNKNOWN = {c:6d} ({100*c/t:2.1f}%)")


def filter_private_mutations(df):
    df.drop(df.columns[df.sum() == 1], inplace=True)



import trisicell as tsc


def deseq():
    try:
        import rpy2.robjects as robjects
        from rpy2.robjects.packages import importr
    except:
        raise SystemExit(
            "A problem was encountered importing `rpy2`. "
            "To run this `rpy2` and `R` need to be installed."
        )
    try:
        importr("Seurat")
    except:
        raise SystemExit(
            "A problem was encountered importing `Seurat` in R. "
            "To run this `Seurat` needs to be installed in R. "
            "Use the following lines to installed them.\n\n"
            "BiocManager::install('Seurat')\n"
        )

    cmd = f"""
    suppressPackageStartupMessages({{
        library(Seurat)
    }})
    """
    robjects.r(cmd)

    # https://satijalab.org/seurat/v3.2/pbmc3k_tutorial.html
    # https://www.rdocumentation.org/packages/Seurat/versions/3.1.4/topics/FindMarkers



import copy
import time
from decimal import *

import numpy as np
import numpy.linalg as la
import pandas as pd
from scipy.special import softmax
from sklearn.metrics.pairwise import pairwise_distances
from tqdm import tqdm


# TODO make this faster by not recalculating
def clt_sample_rec(
    P,
    greedy,
    c,
    names=None,
    namecount=None,
    coef=2,
    prior_prob=Decimal(1.0),
    join_prob_matrix=None,
):
    """
    O(n^2 m^2)

    n: depth of the recursion two edges at each call
    nm^2 : runtime of each call

    can be improved to
    :param P: probability matrix
    :param greedy: sample or max
    :param c: gaussian kernel parameter
    :param names: for rec
    :param namecount: for rec
    :param coef: coef between dist and common mut
    :param prior_prob: for rec
    :return: edges, prior_prob
    """
    if P.shape[0] == 1:
        return list(), prior_prob
    if names is None:
        names = list(range(P.shape[0]))
        namecount = P.shape[0]

    if join_prob_matrix is None:

        def join_neg_priority(a, b):  # smaller values should be joined first
            return la.norm(a - b) - row_leafness_score(a, b) * coef

        dist = pairwise_distances(P, metric=join_neg_priority)  # O(n m^2)
        dist = dist.astype(np.float128)
        np.fill_diagonal(dist, np.inf)

        # This block adjusts c if dist/2c is too big for sotfmax.
        c_rec = c
        for i in range(10):
            sim = softmax(-dist / (2 * c_rec))
            if not np.any(np.isnan(sim)):
                break
            c_rec *= 2.0
    else:
        # add new row to join_prob_matrix to get sim
        pass
    prob = sim
    if greedy:
        pair = np.unravel_index(np.argmax(prob), prob.shape)
    else:
        flat_probs = np.float64(prob.flat)
        ind = np.random.choice(len(flat_probs), p=flat_probs)
        pair = np.unravel_index(ind, prob.shape)

    #  conversion from numpy.float128 to Decimal is not supported
    prior_prob = prior_prob * Decimal(np.float64(prob[pair]))
    P_new = np.delete(P, pair, axis=0)  # remove two rows
    P_new = np.append(
        P_new, np.minimum(P[pair[0]], P[pair[1]]).reshape(1, -1), axis=0
    )  # add one row that has only common ones
    new_edge = [namecount, names[pair[0]], names[pair[1]]]

    new_names = copy.copy(names)
    del new_names[np.max(pair)]
    del new_names[np.min(pair)]
    new_names.append(namecount)
    newnamecount = namecount + 1
    edges, prior_prob = clt_sample_rec(
        P_new, greedy, c, new_names, newnamecount, coef, prior_prob
    )
    edges.append(new_edge)
    return edges, prior_prob


def draw_sample_clt(P, greedy, c=1, coef=2):
    """
    :param P: 
    :param greedy: 
    :param c: gaussian kernel parameter
    :param coef: 
    :return: edges, subtrees, prior_prob
    prior_prob in the latex: Prob_{T\sim E}[T]
    """ ""
    edges, prior_prob = clt_sample_rec(P, greedy, c, coef=coef)
    n_cells = P.shape[0]
    n_nodes = 2 * n_cells - 1
    edges_map = {a: (b, d) for a, b, d in edges}
    subtrees = []
    for i in range(n_nodes):
        if i not in edges_map:  # leaf
            row = np.zeros(n_cells, dtype=np.int8)
            row[i] = 1
        else:
            pair = edges_map[i]
            row = subtrees[pair[0]] + subtrees[pair[1]]  # logical_or
        subtrees.append(row)
    return edges, subtrees, prior_prob


def row_leafness_score(row_a, row_b):
    """
    :return:
    """
    return np.sum(np.minimum(row_a, row_b))


def column_pairs_cost(A, Ap, unit_costs):
    """
    :param A:
    :param Ap:
    :param unit_costs: 2D-matrix: cost of each combination
    :return:
    """
    num = np.zeros((2, 2), dtype=np.int)
    for i in range(2):
        for j in range(2):
            num[i, j] = np.count_nonzero(np.logical_and(A == i, Ap == j))
    return np.sum(num * unit_costs)


def denoise_cond_clt(I, alpha, beta, subtrees):
    """
    Gives a PP matrix with highest likelihood for the given  cell lineage tree.
    O(n m^2) Can be improved to O(n m) with dynamic programming (e.g., in Scistree)
    :param I:
    :param alpha:
    :param beta:
    :param subtrees:
    :return:
    """
    unit_prob = np.array([[1 - alpha, alpha], [beta, 1 - beta]])
    unit_costs = -np.log(unit_prob)
    output = np.zeros(I.shape)
    total_cost = 0
    for c in range(I.shape[1]):
        costs = [
            column_pairs_cost(I[:, c], subtrees[st_ind], unit_costs)
            for st_ind in range(len(subtrees))
        ]
        ind = np.argmin(costs)
        # print(c, costs, ind)
        output[:, c] = subtrees[ind]
        total_cost += costs[ind]
    return output, total_cost



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Farid Rashidi Mehrabadi All rights reserved.

# =========================================================================================
# Author     : Farid Rashidi Mehrabadi (farid.rashidimehrabadi@nih.gov)
# Last Update: Jan 1, 2021
# Description: get the expression using RSEM
# =========================================================================================

from trisicell.ul._servers import *


def run09(config, afterok):
    def cmds(sample):
        cmds = ""
        cmds += cmd([f"mkdir -p {config['outdir']}/{sample}"])
        if config["pairedend"]:
            cmds += cmd(
                [
                    f"rsem-calculate-expression",
                    f"--bam",
                    f"--no-bam-output",
                    f"--paired-end",
                    f"--estimate-rspd",
                    f"--append-names",
                    f"{config['outdir']}/{sample}/Aligned.toTranscriptome.out.bam",
                    f"{config['rsemdb']}",
                    f"{config['outdir']}/{sample}/expr",
                ]
            )
        else:
            cmds += cmd(
                [
                    f"rsem-calculate-expression",
                    f"--bam",
                    f"--no-bam-output",
                    f"--estimate-rspd",
                    f"--append-names",
                    f"{config['outdir']}/{sample}/Aligned.toTranscriptome.out.bam",
                    f"{config['rsemdb']}",
                    f"{config['outdir']}/{sample}/expr",
                ]
            )
        cmds += cmd(
            [
                f"rm -rf",
                f"{config['outdir']}/{sample}/expr.stat",
                f"{config['outdir']}/{sample}/expr.isoforms.results",
            ]
        )
        cmds += cmd([f"echo Done!"], islast=True)
        return cmds

    df = pd.DataFrame()
    df["sample"] = config["samples"]
    df["cmd"] = df.apply(lambda x: cmds(x["sample"]), axis=1)

    jobname = os.path.basename(__file__)[:-3]
    cmdmain = write_cmds_get_main(
        df,
        jobname,
        config["time09"],
        config["mem09"],
        "STAR/2.7.3a,rsem/1.3.2",
        1,
        config["email"],
        config["tmpdir"],
        afterok,
    )
    return cmdmain



import os
import time

import numpy as np
import pandas as pd

import trisicell as tsc


def onconem(df_input, alpha, beta):
    """Running OncoNEM.

    Parameters
    ----------
    df_input : :class:`pandas.DataFrame`
        Input genotype matrix where 0 is absent, 1 is present and 3 is missing.
    alpha : :obj:`float`
        False positive rate.
    beta : :obj:`float`
        False negative rate.

    Returns
    -------
    :class:`pandas.DataFrame`
        Conflict-free genotype matrix where 0 is absent, 1 is present.
    """

    try:
        import rpy2.robjects as robjects
        from rpy2.robjects.packages import importr
    except:
        raise SystemExit(
            "A problem was encountered importing `rpy2`. "
            "To run this `rpy2` and `R` need to be installed."
        )
    try:
        importr("oncoNEM")
    except:
        raise SystemExit(
            "A problem was encountered importing `oncoNEM` in R. "
            "To run this `oncoNEM` needs to be installed in R. "
            "Use the following lines to installed them.\n\n"
            "BiocManager::install('graph')\n"
            "devtools::install_bitbucket('edith_ross/oncoNEM')\n"
        )

    tsc.logg.info(f"running OncoNEM with alpha={alpha}, beta={beta}")

    robjects.pandas2ri.activate()
    df_r = robjects.conversion.py2rpy(df_input.replace(3, 2).T)
    robjects.globalenv["df"] = df_r

    cmd = f"""
    suppressPackageStartupMessages({{
        library(oncoNEM)
        library(igraph)
    }})
    
    mat <- data.matrix(df)
    oNEM <- oncoNEM$new(Data=mat, FPR=as.numeric({alpha}), FNR=as.numeric({beta}))
    oNEM$search(delta=200)
    oNEM.expanded <- expandOncoNEM(oNEM, epsilon=10, delta=200,
                                   checkMax=10000, app=TRUE)
    oncoTree <-clusterOncoNEM(oNEM=oNEM.expanded, epsilon=10)
    edges <- get.edgelist(oncoTree$g)
    post <- oncoNEMposteriors(tree=oncoTree$g, clones=oncoTree$clones,
                              Data=oNEM$Data, FPR=oNEM$FPR, FNR=oNEM$FNR)
    
    post$p_mut
    """
    s_time = time.time()
    result = robjects.r(cmd)
    e_time = time.time()
    running_time = e_time - s_time

    sol_Y = np.rint(result)

    df_output = pd.DataFrame(sol_Y)
    df_output = df_output.T
    df_output.columns = df_input.columns
    df_output.index = df_input.index
    df_output.index.name = "cellIDxmutID"
    df_output = df_output.astype(int)

    tsc.ul.stat(df_input, df_output, alpha, beta, running_time)

    return df_output



import itertools

import networkx as nx
import numpy as np
import pandas as pd

import trisicell as tsc


def consensus_combine(df):
    """Combine the replicates/cells in genotype matrix.

    This function combine the genotype of replicates/cells that are
    in the form of `{CellA}_{id}` to `{CellA}`.

    Parameters
    ----------
    df : :class:`pandas.DataFrame`
        The input genotype matrix in conflict-free format.

    Returns
    -------
    :class:`pandas.DataFrame`
        The combine genotype matrix.
    """

    df2 = df.groupby(df.index.str.split("_").str[0]).transform("prod")
    df2 = df2.groupby(df2.index.str.split("_").str[0]).first()
    return df2


def _consensus_of_cfmatrices(df1, df2):
    # TODO: unused
    df1_temp = df1.loc[df1.index]
    df2_temp = df2.loc[df1.index]
    mut1 = pd.DataFrame(df1.columns, columns=["muts"])
    mut1[["ensemble", "gene", "chrom", "position", "reference", "alteration"]] = (
        mut1["muts"].apply(tsc.ul.split_mut).apply(pd.Series)
    )
    mut2 = pd.DataFrame(df2.columns, columns=["muts"])
    mut2[["ensemble", "gene", "chrom", "position", "reference", "alteration"]] = (
        mut2["muts"].apply(tsc.ul.split_mut).apply(pd.Series)
    )
    mut1 = mut1.set_index("muts")
    mut2 = mut2.set_index("muts")
    mut1_t = (
        mut1["chrom"].astype(str)
        + ":"
        + mut1["position"].astype(str)
        + ":"
        + mut1["reference"].astype(str)
        + ":"
        + mut1["alteration"].astype(str)
    )
    mut2_t = (
        mut2["chrom"].astype(str)
        + ":"
        + mut2["position"].astype(str)
        + ":"
        + mut2["reference"].astype(str)
        + ":"
        + mut2["alteration"].astype(str)
    )
    mut1_t = pd.Series(mut1_t.index, index=mut1_t)
    mut2_t = pd.Series(mut2_t.index, index=mut2_t)
    shared = np.intersect1d(mut1_t.index, mut2_t.index)
    finals = []
    for mut1, mut2 in zip(mut1_t[shared].values, mut2_t[shared].values):
        if df1_temp[mut1].equals(df2_temp[mut2]):
            finals.append(mut1)
    return df1[finals]


def _get_cnt_tree(tree):
    cnt_tree = tree.copy()
    for u, v, l in cnt_tree.edges.data("label"):
        mutations = l.split(cnt_tree.graph["splitter_mut"])
        cnt_tree.add_edge(u, v, label=len(mutations), mutations=mutations)
        if cnt_tree.in_degree(v) != 0:
            if "––" not in tree.nodes[v]["label"]:
                cnt_tree.nodes[v]["label"] = cnt_tree.nodes[v]["label"].split(
                    cnt_tree.graph["splitter_cell"]
                )
    return cnt_tree


def _get_node_labels(tree):
    nl = {}
    for x in tree.nodes:
        if tree.in_degree(x) != 0:
            if "––" not in tree.nodes[x]["label"]:
                for y in tree.nodes[x]["label"]:
                    nl[y] = x
        else:
            nl["root"] = x
    return nl


def _has_path(tree, nodes, l1, l2):
    try:
        n1 = nodes[l1]
        n2 = nodes[l2]
        dij = nx.bidirectional_dijkstra(tree, n1, n2, weight="label")
        return True, dij[0]
    except:
        return False, None
    return False, None


def _merge_with_parents(tree, nodes, n1, n2):
    lca = nx.lowest_common_ancestor(tree, n1, n2)
    costs, paths = nx.bidirectional_dijkstra(tree, lca, n1, weight="label")
    paths = paths[::-1]
    labels = []
    for p in paths[1:]:
        #### merge-with-parent, n1 -> n2 (n2 is going to be replaced by n1)
        if "––" not in tree.nodes[p]["label"]:
            if "––" not in tree.nodes[n1]["label"]:
                tree.nodes[n1]["label"] += tree.nodes[p]["label"]
            else:
                tree.nodes[n1]["label"] = tree.nodes[p]["label"]
            for x in tree.nodes[p]["label"]:
                nodes[x] = n1
        cost = tree.edges[p, n1]["label"]

        #### add the mutations of this edge to the grandparent edge
        grandparent = list(tree.predecessors(p))[0]
        tree.edges[(grandparent, p)]["mutations"] += tree.edges[(p, n1)]["mutations"]
        tree.edges[(grandparent, p)]["label"] += tree.edges[(p, n1)]["label"]

        tree = nx.contracted_nodes(tree, n1, p, self_loops=False)
    return tree, nodes, costs


def _get_labels(cnt_tree):
    labels = {}
    [y for y in cnt_tree.nodes if "––" in cnt_tree.nodes[y]["label"]]
    for x in cnt_tree.nodes:
        if "root" not in cnt_tree.nodes[x]["label"]:
            lbs = []
            sub = nx.subgraph(
                cnt_tree,
                nx.algorithms.traversal.depth_first_search.dfs_tree(cnt_tree, x).nodes,
            )
            for y in sub.nodes:
                if "––" not in cnt_tree.nodes[y]["label"]:
                    lbs += cnt_tree.nodes[y]["label"]
            labels[x] = lbs
    return labels


def _add_private_muts(cnt_tree, sc_data, tree_nodes):
    latest = max(cnt_tree.nodes)
    nodes = list(cnt_tree.nodes)
    has_expanded_all = []
    for node in nodes:
        if "––" not in cnt_tree.nodes[node]["label"] and cnt_tree.in_degree(node) != 0:
            has_expanded = []
            for cell in cnt_tree.nodes[node]["label"]:
                private = (sc_data.loc[cell] == 1) & (
                    sc_data.loc[np.setdiff1d(sc_data.index, cell)].sum() == 0
                )
                private = list(private[private == True].index)
                if len(private) > 0:
                    has_expanded_all.append(cell)
                    has_expanded.append(cell)
                    latest = latest + 1
                    cnt_tree.add_node(latest, label=[cell])
                    cnt_tree.add_edge(
                        node, latest, label=len(private), mutations=private
                    )
                    tree_nodes[cell] = latest
            for cell in has_expanded:
                cnt_tree.nodes[node]["label"].remove(cell)
            if len(cnt_tree.nodes[node]["label"]) == 0:
                cnt_tree.nodes[node]["label"] = "––"
    return cnt_tree, tree_nodes, has_expanded_all


def consensus_tree(sc1, sc2):
    """Building the consensus tree between to phylogenetic trees.

    Parameters
    ----------
    df1 : :class:`pandas.DataFrame`
        First conflict-free matrix.
    df2 : :class:`pandas.DataFrame`
        Second conflict-free matrix.

    Returns
    -------
    :class:`networkx.DiGraph`
        Two trees derived by building the consensus procedure.
    """

    #### remove private mutations and build the trees
    sc1_c = sc1.loc[:, sc1.sum() > 1].copy()
    sc2_c = sc2.loc[:, sc2.sum() > 1].copy()

    tree1 = tsc.ul.to_tree(sc1_c)
    tree2 = tsc.ul.to_tree(sc2_c)

    cnt_tree1 = _get_cnt_tree(tree1)
    cnt_tree2 = _get_cnt_tree(tree2)

    nodes1 = _get_node_labels(cnt_tree1)
    nodes2 = _get_node_labels(cnt_tree2)

    common_cells = list(np.intersect1d(nodes1.keys(), nodes2.keys())[0])

    total_cost = 0

    #### step 1,2
    for x, y in itertools.permutations(common_cells, 2):
        hp1, cost1 = _has_path(cnt_tree1, nodes1, x, y)
        if hp1:
            hp2, cost2 = _has_path(cnt_tree2, nodes2, x, y)
            if not hp2:
                cnt_tree2, nodes2, cost2 = _merge_with_parents(
                    cnt_tree2, nodes2, nodes2[x], nodes2[y]
                )
                tsc.logg.info(f"merge in tree2: `{x}` to `{y}` and cost={cost2}")
                total_cost += cost2

    for x, y in itertools.permutations(common_cells, 2):
        hp2, cost2 = _has_path(cnt_tree2, nodes2, x, y)
        if hp2:
            hp1, cost1 = _has_path(cnt_tree1, nodes1, x, y)
            if not hp1:
                cnt_tree1, nodes1, cost1 = _merge_with_parents(
                    cnt_tree1, nodes1, nodes1[x], nodes1[y]
                )
                tsc.logg.info(f"merge in tree1: `{x}` to `{y}` and cost={cost1}")
                total_cost += cost1

    for x, y in itertools.permutations(common_cells, 2):
        hp1, cost1 = _has_path(cnt_tree1, nodes1, x, y)
        if hp1:
            hp2, cost2 = _has_path(cnt_tree2, nodes2, x, y)
            if not hp2:
                cnt_tree2, nodes2, cost2 = _merge_with_parents(
                    cnt_tree2, nodes2, nodes2[x], nodes2[y]
                )
                tsc.logg.info(f"merge in tree2: `{x}` to `{y}` and cost={cost2}")
                total_cost += cost2

    #### step 3
    labels1 = _get_labels(cnt_tree1)
    labels2 = _get_labels(cnt_tree2)

    matched = []
    for x, y in labels1.items():
        for p, q in labels2.items():
            if set(y) == set(q):
                matched.append((x, p))
    for i, j in matched:
        del labels1[i]
        del labels2[j]

    for y, _ in labels1.items():
        parent = list(cnt_tree1.predecessors(y))[0]
        cnt_tree1, nodes1, cost1 = _merge_with_parents(cnt_tree1, nodes1, y, parent)
        tsc.logg.info(
            f"merge in tree1 - internal: `node{y}` to `node{parent}` and cost={cost1}"
        )
        total_cost += cost1

    for y, _ in labels2.items():
        parent = list(cnt_tree2.predecessors(y))[0]
        cnt_tree2, nodes2, cost2 = _merge_with_parents(cnt_tree2, nodes2, y, parent)
        tsc.logg.info(
            f"merge in tree2 - internal: `node{y}` to `node{parent}` and cost={cost2}"
        )
        total_cost += cost2

    #### add private mutations back
    cnt_tree1, nodes1, has_expanded1 = _add_private_muts(cnt_tree1, sc1, nodes1)
    cnt_tree2, nodes2, has_expanded2 = _add_private_muts(cnt_tree2, sc2, nodes2)

    for y in np.setdiff1d(has_expanded1, has_expanded2):
        parent = list(cnt_tree1.predecessors(nodes1[y]))[0]
        cnt_tree1, nodes1, cost1 = _merge_with_parents(
            cnt_tree1, nodes1, nodes1[y], parent
        )
        tsc.logg.info(
            f"merge in tree1 - private: `{y}` to `node{parent}` and cost={cost1}"
        )
        total_cost += cost1

    for y in np.setdiff1d(has_expanded2, has_expanded1):
        parent = list(cnt_tree2.predecessors(nodes2[y]))[0]
        cnt_tree2, nodes2, cost2 = _merge_with_parents(
            cnt_tree2, nodes2, nodes2[y], parent
        )
        tsc.logg.info(
            f"merge in tree2 - private: `{y}` to `node{parent}` and cost={cost2}"
        )
        total_cost += cost1

    tsc.logg.info("    ---> total cost:", total_cost)

    if nx.is_isomorphic(cnt_tree1, cnt_tree2) == False:
        raise RuntimeError("Error: Two trees are not isomorphic!")

    final_tree = cnt_tree1.copy()
    root1 = [x for x in cnt_tree1.nodes if cnt_tree1.in_degree(x) == 0][0]
    root2 = [x for x in cnt_tree2.nodes if cnt_tree2.in_degree(x) == 0][0]
    leaves2 = [x for x in cnt_tree2.nodes if cnt_tree2.out_degree(x) == 0]
    for leaf2 in leaves2:
        list2 = nx.dijkstra_path(cnt_tree2, root2, leaf2)
        label1 = cnt_tree2.nodes[leaf2]["label"]

        mutations2 = []
        for first, second in zip(list2, list2[1:]):
            mutations2.append(cnt_tree2.edges[(first, second)]["mutations"])

        list1 = nx.dijkstra_path(cnt_tree1, root1, nodes1[label1[0]])
        i = 0
        for first, second in zip(list1, list1[1:]):
            mutations1 = cnt_tree1.edges[(first, second)]["mutations"]
            mutations = np.intersect1d(mutations1, mutations2[i])
            final_tree.add_edge(
                first, second, label=final_tree.graph["splitter_mut"].join(mutations)
            )
            i += 1
    for v in final_tree.nodes:
        if final_tree.in_degree(v) != 0:
            if "––" not in final_tree.nodes[v]["label"]:
                final_tree.nodes[v]["label"] = final_tree.graph["splitter_cell"].join(
                    final_tree.nodes[v]["label"]
                )

    data = tsc.ul.to_cfmatrix(final_tree)
    final_tree.graph["data"] = data
    del final_tree.graph["become_germline"]

    return final_tree



from trisicell.tl.solver._bnb import bnb
from trisicell.tl.solver._cardelino import cardelino
from trisicell.tl.solver._dendro import dendro
from trisicell.tl.solver._onconem import onconem
from trisicell.tl.solver._phiscs import phiscs_original, phiscsb, phiscsi
from trisicell.tl.solver._scistree import iscistree, rscistree, scistree
from trisicell.tl.solver._scite import infscite, scite
from trisicell.tl.solver._siclonefit import siclonefit
from trisicell.tl.solver.booster import booster
from trisicell.tl.solver.huntress import huntress



from operator import index

import networkx as nx
import numpy as np
import pandas as pd

import trisicell as tsc


def to_cfmatrix(tree):
    # TODO: implement.
    pass


def to_tree(df):
    """Convert a conflict-free matrix to a tree object.

    This function converts a conflict-free matrix to a tree object in which
    nodes are labled with cells and edges are lables with mutations. The root is
    labled by 'root'. Mutations are seperated by `.graph['splitter_mut']` and cells
    are seperated by `.graph['splitter_cell']`. Those mutations that are not present
    in any cell are stored in `.graph['become_germline']`. Mutations happed once
    during the evolution so there is no repetitive mutation.

    Parameters
    ----------
    df : :class:`pandas.DataFrame`
        A genotype dataframe in which rows are cells and columns are mutations.
        Note that this dataframe must be conflict-free.

    Returns
    -------
    :class:`networkx.DiGraph`
        A perfect phylogenetic tree.

    Raises
    ------
    Exception
        If the input dataframe is not conflict-free.
    """

    if not tsc.ul.is_conflict_free_gusfield(df):
        raise Exception("The input is not conflict-free!")

    def _contains(col1, col2):
        for i in range(len(col1)):
            if not col1[i] >= col2[i]:
                return False
        return True

    tree = nx.DiGraph()
    tree.graph["data"] = df
    tree.graph["splitter_mut"] = "\n"
    tree.graph["splitter_cell"] = "\n"
    tree.graph["become_germline"] = df.columns[(df == 0).all(axis=0)]

    matrix = df.values
    names_mut = list(df.columns)

    i = 0
    while i < matrix.shape[1]:
        j = i + 1
        while j < matrix.shape[1]:
            if np.array_equal(matrix[:, i], matrix[:, j]):
                matrix = np.delete(matrix, j, 1)
                x = names_mut.pop(j)
                names_mut[i] += tree.graph["splitter_mut"] + x
                j -= 1
            j += 1
        i += 1

    rows = matrix.shape[0]
    cols = matrix.shape[1]
    dimensions = np.sum(matrix, axis=0)
    indices = np.argsort(dimensions)
    dimensions = np.sort(dimensions)
    names_mut = [names_mut[indices[i]] for i in range(cols)]

    tree.add_node(cols)
    tree.add_node(cols - 1)
    tree.add_edge(cols, cols - 1, label=names_mut[cols - 1])
    node_mud = {}
    node_mud[names_mut[cols - 1]] = cols - 1

    i = cols - 2
    while i >= 0:
        if dimensions[i] == 0:
            break
        attached = False
        for j in range(i + 1, cols):
            if _contains(matrix[:, indices[j]], matrix[:, indices[i]]):
                tree.add_node(i)
                tree.add_edge(node_mud[names_mut[j]], i, label=names_mut[i])
                node_mud[names_mut[i]] = i
                attached = True
                break
        if not attached:
            tree.add_node(i)
            tree.add_edge(cols, i, label=names_mut[i])
            node_mud[names_mut[i]] = i
        i -= 1

    clusters = {cols: "root"}
    for node in tree:
        if node == cols:
            tree.nodes[node]["label"] = "root"
            continue
        untilnow_mut = []
        sp = nx.shortest_path(tree, cols, node)
        for i in range(len(sp) - 1):
            untilnow_mut += tree.get_edge_data(sp[i], sp[i + 1])["label"].split(
                tree.graph["splitter_mut"]
            )
        untilnow_cell = df.loc[
            (df[untilnow_mut] == 1).all(axis=1)
            & (df[[x for x in df.columns if x not in untilnow_mut]] == 0).all(axis=1)
        ].index
        if len(untilnow_cell) > 0:
            clusters[node] = f"{tree.graph['splitter_cell'].join(untilnow_cell)}"
        else:
            clusters[node] = "––"

        tree.nodes[node]["label"] = clusters[node]

    i = 1
    for k, v in clusters.items():
        if v == "––":
            clusters[k] = i * "––"
            i += 1
    return tree


def to_cfmatrix(tree):
    """Convert phylogenetic tree to conflict-free matrix.

    This function converts a phylogenetic tree in which mutations
    are at edges and cells are at nodes to a conflict-free matrix
    where rows are cells, columns are mutations and each entry is
    either zero or one representing the absence or presence of the
    mutation in the cell.

    Parameters
    ----------
    tree : :class:`networkx.DiGraph`
        The phylogenetic tree.

    Returns
    -------
    :class:`pandas.DataFrame`
        The conflict-free matrix.
    """

    mutations = []
    cells = []
    for _, v, l in tree.edges(data=True):
        mutations += l["label"].split(tree.graph["splitter_mut"])
        if "––" not in tree.nodes[v]["label"]:
            cells += tree.nodes[v]["label"].split(tree.graph["splitter_cell"])
    df = pd.DataFrame(0, index=cells, columns=mutations)
    root = [x for x in tree.nodes if tree.in_degree(x) == 0][0]
    leaves = [x for x in tree.nodes if tree.out_degree(x) == 0]
    for leaf in leaves:
        nodes = nx.dijkstra_path(tree, root, leaf)
        mut = []
        for first, second in zip(nodes, nodes[1:]):
            mut += tree.edges[(first, second)]["label"].split(
                tree.graph["splitter_mut"]
            )
            if "––" not in tree.nodes[second]["label"]:
                cell = tree.nodes[second]["label"].split(tree.graph["splitter_cell"])
                df.loc[cell, mut] = 1
    return df


def _to_mutation_tree(tree):
    tree2 = nx.DiGraph()
    for u, v, l in tree.edges.data("label"):
        if tree.in_degree(u) == 0:
            tree2.add_node(u, label="root")
        muts = l.split(tree.graph["splitter_mut"])
        tree2.add_node(v, label=muts)
        tree2.add_edge(u, v)
    return tree2


def _to_newick(tree):
    def subtree(at):
        return nx.subgraph(
            tree,
            nx.algorithms.traversal.depth_first_search.dfs_tree(tree, at).nodes - [at],
        )

    def children(at):
        return [n for n in tree.neighbors(at)]

    root = [x for x in tree.nodes if tree.in_degree(x) == 0][0]

    def newick_recursive(node_id):
        node_ids = children(node_id)
        if len(node_ids) == 0:
            cells = tree.nodes[node_id]["label"].split(tree.graph["splitter_cell"])
            return "(" + ":1,".join(cells) + f":1)Node{node_id+1}:1"
        elif len(node_ids) > 0:
            cells = tree.nodes[node_id]["label"].split(tree.graph["splitter_cell"])
            if not ("––" in cells or "root" in cells):
                return (
                    "("
                    + ":1,".join(cells)
                    + ":1,"
                    + ",".join([newick_recursive(node_id) for node_id in node_ids])
                    + f":1)Node{node_id+1}"
                )
            else:
                return (
                    "("
                    + ":1,".join([newick_recursive(node_id) for node_id in node_ids])
                    + f":1)Node{node_id+1}:1"
                )

    newick = newick_recursive(root) + ";"
    return newick


def _info2_mutation_list(tree):
    tree2 = _to_mutation_tree(tree)

    row = []
    for node in tree2.nodes:
        if tree2.in_degree(node) == 0:
            row.append(
                {
                    "newick_label": f"Node{node+1}",
                    "nmuts_label": "root",
                    "nodeid_label": f"[{node+1}]",
                    "mixed_label": f"root – [{node+1}]",
                }
            )
        else:
            row.append(
                {
                    "newick_label": f"Node{node+1}",
                    "nmuts_label": f"{len(tree2.nodes[node]['label'])}",
                    "nodeid_label": f"[{node+1}]",
                    "mixed_label": (f"{len(tree2.nodes[node]['label'])} – [{node+1}]"),
                }
            )
    info2 = pd.DataFrame(row)

    row = []
    for node in tree2.nodes:
        if tree2.in_degree(node) == 0:
            continue
        for mut in tree2.nodes[node]["label"]:
            ens, gene, chrom, pos, ref, alt = tsc.ul.split_mut(mut)
            row.append(
                {
                    "Node": f"[{node+1}]",
                    "Ensemble": ens,
                    "Gene": gene,
                    "Chrom": chrom,
                    "Position": pos,
                    "Reference": ref,
                    "Alteration": alt,
                }
            )
    mutation_list = pd.DataFrame(row)
    return info2, mutation_list


def _split_labels(mt, mt_guide):
    root_guide = [node for node in mt_guide.nodes if mt_guide.in_degree(node) == 0][0]
    guide = [
        y
        for x in nx.algorithms.bfs_tree(mt_guide, root_guide).nodes
        for y in mt_guide.nodes[x]["label"]
    ]

    root = [node for node in mt.nodes if mt.in_degree(node) == 0][0]
    nodes = nx.algorithms.bfs_tree(mt, root).nodes

    latest_node = root
    removing = []
    for node in nodes:
        muts = mt.nodes[node]["label"]
        if len(muts) > 1 and node != root:
            parent = list(mt.predecessors(node))[-1]
            if mt.out_degree(node) != 0:
                children = list(mt.successors(node))
            else:
                children = -1
            muts = sorted(muts, key=lambda i: guide.index(i))

            removing.append(node)
            for i, x in enumerate(muts):
                latest_node += 1
                mt.add_node(latest_node, label=x)
                if i == 0:
                    mt.add_edge(parent, latest_node)
                elif i == len(muts) - 1:
                    if children != -1:
                        mt.add_edge(latest_node - 1, latest_node)
                        for child in children:
                            mt.add_edge(latest_node, child)
                    else:
                        mt.add_edge(latest_node - 1, latest_node)
                else:
                    mt.add_edge(latest_node - 1, latest_node)
        else:
            if node != root:
                mt.nodes[node]["label"] = mt.nodes[node]["label"][0]

    for node in removing:
        mt.remove_node(node)
    return mt


def _to_apted(sl_tree):
    def children(at):
        return [n for n in sl_tree.neighbors(at)]

    def apted_recursive(node):
        nodes = children(node)
        if len(nodes) == 0:
            l = sl_tree.nodes[node]["label"]
            return "{" + l + "}"
        else:
            l = sl_tree.nodes[node]["label"]
            x = ""
            for node in nodes:
                x += apted_recursive(node)
            return "{" + l + x + "}"

    root = [node for node in sl_tree.nodes if sl_tree.in_degree(node) == 0][0]
    return apted_recursive(root)


def iterate(tree):
    for (u, v, e) in tree.edges.data():
        yield u.split(tree.graph["splitter_cell"]), v.split(
            tree.graph["splitter_cell"]
        ), e["label"].split(tree.graph["splitter_mut"])


def partition_cells(tree, node):
    nd = tree.graph["splitter_cell"].join(node)
    cells = []
    for x in list(nx.algorithms.traversal.depth_first_search.dfs_tree(tree, nd).nodes):
        for y in x.split(", "):
            if not "–" in y:
                cells.append(y)
    cells = np.array(cells)
    return cells, np.setdiff1d(tree.graph["data"].index, cells)


def cells_rooted_at(tree, node_id):
    df = tree.graph["mutation_list"][tree.graph["mutation_list"].Node == node_id]
    cells = (tree.graph["data"][df.index] == 1).all(axis=1)
    cells = np.array(tree.graph["data"].loc[cells].index)
    return cells, np.setdiff1d(tree.graph["data"].index, cells)


def muts_rooted_at(tree):
    muts = tree.graph["mutation_list"][
        tree.graph["mutation_list"].Node == node_id
    ].index
    return np.array(muts)
